from mock import Mock, call, patch
import json
from oauth2.client_authenticator import ClientAuthenticator
from oauth2.compatibility import quote
from oauth2.test import unittest
from oauth2.web import Response, ResourceOwnerGrantSiteAdapter, \
    ImplicitGrantSiteAdapter, AuthorizationCodeGrantSiteAdapter
from oauth2.web.wsgi import Request
from oauth2.grant import ImplicitGrantHandler, AuthorizationCodeAuthHandler, \
    AuthRequestMixin, AuthorizationCodeTokenHandler, ImplicitGrant, \
    AuthorizationCodeGrant, ResourceOwnerGrantHandler, ResourceOwnerGrant, \
    Scope, RefreshToken, RefreshTokenHandler, ClientCredentialsGrant, \
    ClientCredentialsHandler, AuthorizeMixin, SiteAdapterMixin
from oauth2.store import ClientStore, AuthCodeStore, AccessTokenStore
from oauth2.error import OAuthInvalidError, UserNotAuthenticated, \
    AccessTokenNotFound, UserIdentifierMissingError, AuthCodeNotFound
from oauth2 import Provider
from oauth2.datatype import Client, AuthorizationCode, AccessToken
from oauth2.tokengenerator import TokenGenerator
from copy import copy


def mock_time():
    return 1000


class AuthorizationCodeGrantTestCase(unittest.TestCase):
    def test_create_auth_handler(self):
        """
        AuthorizationCodeGrant() should return a new instance of AuthorizationCodeAuthHandler on request
        """
        default_scope = "default_scope"
        method = "GET"
        scopes = ["first", "second"]
        path = "/auth"

        request_mock = Mock(spec=Request)
        request_mock.method = method
        request_mock.path = path
        request_mock.get_param.return_value = "code"

        scope_mock = Mock(Scope)

        server_mock = Mock()
        server_mock.authorize_path = path
        server_mock.auth_code_store = Mock()
        server_mock.client_authenticator = Mock()
        server_mock.token_generator = Mock()

        factory = AuthorizationCodeGrant(
            default_scope=default_scope,
            scopes=scopes,
            scope_class=scope_mock,
            site_adapter=Mock(spec=AuthorizationCodeGrantSiteAdapter)
        )
        result_class = factory(request_mock, server_mock)

        request_mock.get_param.assert_called_with("response_type")
        scope_mock.assert_called_with(default=default_scope, available=scopes)
        self.assertTrue(isinstance(result_class, AuthorizationCodeAuthHandler))

    def test_create_token_handler(self):
        path = "/token"

        request_mock = Mock(spec=Request)
        request_mock.method = "POST"
        request_mock.path = path
        request_mock.post_param.return_value = "authorization_code"

        server_mock = Mock()
        server_mock.authorize_path = "/auth"
        server_mock.token_path = path
        server_mock.access_token_store = Mock(spec=AccessTokenStore)
        server_mock.auth_code_store = Mock()
        server_mock.client_store = Mock()
        server_mock.token_generator = Mock()

        factory = AuthorizationCodeGrant(
            site_adapter=Mock(spec=AuthorizationCodeGrantSiteAdapter)
        )
        result_class = factory(request_mock, server_mock)

        request_mock.post_param.assert_called_with("grant_type")
        self.assertTrue(isinstance(result_class,
                                   AuthorizationCodeTokenHandler))

    def test_create_no_match(self):
        request_mock = Mock(spec=Request)
        request_mock.method = "POST"
        request_mock.get_param.return_value = "no-code"
        request_mock.post_param.return_value = "no-authorization_code"

        factory = AuthorizationCodeGrant(
            site_adapter=Mock(spec=AuthorizationCodeGrantSiteAdapter)
        )
        result_class = factory(request_mock, Mock())

        request_mock.post_param.assert_called_with("grant_type")
        self.assertEqual(result_class, None)


class AuthRequestMixinTestCase(unittest.TestCase):
    def test_read_validate_params_all_valid(self):
        """
        AuthRequestMixin.read_validate_params should parse all params correctly if they are valid
        """
        response_type = "code"
        state = "state"

        client = Client(identifier="abc", secret="xyz",
                        redirect_uris=["http://callback"],
                        authorized_response_types=["code"])

        request_mock = Mock(spec=Request)
        request_mock.get_param.side_effect = [response_type, state]

        scope_handler_mock = Mock(Scope)

        client_auth_mock = Mock(spec=ClientAuthenticator)
        client_auth_mock.by_identifier.return_value = client

        handler = AuthRequestMixin(client_authenticator=client_auth_mock,
                                   scope_handler=scope_handler_mock,
                                   token_generator=Mock())

        result = handler.read_validate_params(request_mock)

        request_mock.get_param.assert_has_calls([call("state")])
        scope_handler_mock.parse.assert_called_with(request_mock, "query")
        client_auth_mock.by_identifier.assert_called_with(request_mock)
        self.assertEqual(handler.state, state)
        self.assertTrue(result)


class AuthorizeMixinTestCase(unittest.TestCase):
    def test_authorize_user_denied_access(self):
        """
        AuthorizeMixin.authorize should raise an OAuthUserError if the user did not authorize the request
        """
        site_adapter_mock = Mock(spec=ImplicitGrantSiteAdapter)
        site_adapter_mock.user_has_denied_access.return_value = True

        auth_mixin = AuthorizeMixin(site_adapter=site_adapter_mock)
        with self.assertRaises(OAuthInvalidError):
            auth_mixin.authorize(Mock(spec=Request), Mock(spec=Response),
                                 environ={}, scopes=[])

    def test_authorize_dict_return(self):
        """
        AuthorizeMixin.authorize should return a tuple even if the SiteAdapter returns a dict
        """
        client_mock = Mock(spec=Client)

        test_data = {"test": "data"}

        site_adapter_mock = Mock(spec=ImplicitGrantSiteAdapter)
        site_adapter_mock.user_has_denied_access.return_value = False
        site_adapter_mock.authenticate.return_value = test_data

        auth_mixin = AuthorizeMixin(site_adapter=site_adapter_mock)
        auth_mixin.client = client_mock
        result = auth_mixin.authorize(Mock(spec=Request), Mock(spec=Response),
                                      environ={}, scopes=[])

        self.assertTrue(isinstance(result, tuple))
        self.assertDictEqual(result[0], test_data)
        self.assertIsNone(result[1])

    def test_authorize_tuple_return(self):
        """
        AuthorizeMixin.authorize should return the tuple returned by the SiteAdapter
        """
        client_mock = Mock(spec=Client)

        test_data = ({"test": "data"}, 123)

        site_adapter_mock = Mock(spec=ImplicitGrantSiteAdapter)
        site_adapter_mock.user_has_denied_access.return_value = False
        site_adapter_mock.authenticate.return_value = test_data

        auth_mixin = AuthorizeMixin(site_adapter=site_adapter_mock)
        auth_mixin.client = client_mock
        result = auth_mixin.authorize(Mock(spec=Request), Mock(spec=Response),
                                      environ={}, scopes=[])

        self.assertTrue(isinstance(result, tuple))
        self.assertDictEqual(result[0], test_data[0])
        self.assertEqual(result[1], test_data[1])

    def test_authorize_user_not_authenticated(self):
        client_mock = Mock(spec=Client)

        response_mock = Mock(spec=Response)

        site_adapter_mock = Mock(spec=ImplicitGrantSiteAdapter)
        site_adapter_mock.user_has_denied_access.return_value = False
        site_adapter_mock.authenticate.side_effect = UserNotAuthenticated
        site_adapter_mock.render_auth_page.return_value = response_mock

        auth_mixin = AuthorizeMixin(site_adapter=site_adapter_mock)
        auth_mixin.client = client_mock
        result = auth_mixin.authorize(Mock(spec=Request), response_mock,
                                      environ={}, scopes=[])

        self.assertEqual(result, response_mock)


class AuthorizationCodeAuthHandlerTestCase(unittest.TestCase):
    def test_process(self):
        code = "abcd"
        environ = {"session": "data"}
        scopes = ["scope"]
        state = "my%state"
        redirect_uri = "https://callback"
        user_data = {"user_id": 789}

        location_uri = "%s?code=%s&state=%s" % (redirect_uri, code, quote(state))

        auth_code_store_mock = Mock(spec=AuthCodeStore)

        client = Client(identifier="abc", secret="xyz",
                        redirect_uris=[redirect_uri])

        response_mock = Mock(spec=Response)

        request_mock = Mock(spec=Request)

        scope_handler_mock = Mock(Scope)
        scope_handler_mock.scopes = scopes
        scope_handler_mock.send_back = False

        site_adapter_mock = Mock(spec=AuthorizationCodeGrantSiteAdapter)
        site_adapter_mock.authenticate.return_value = user_data
        site_adapter_mock.user_has_denied_access.return_value = False

        token_generator_mock = Mock(spec=["generate"])
        token_generator_mock.generate.return_value = code

        handler = AuthorizationCodeAuthHandler(
            auth_token_store=auth_code_store_mock,
            client_authenticator=Mock(), scope_handler=scope_handler_mock,
            site_adapter=site_adapter_mock,
            token_generator=token_generator_mock
        )

        handler.client = client
        handler.state = state
        response = handler.process(request_mock, response_mock, environ)

        token_generator_mock.generate.assert_called_with()
        site_adapter_mock.authenticate.assert_called_with(request_mock,
                                                          environ, scopes,
                                                          client)
        self.assertTrue(auth_code_store_mock.save_code.called)
        self.assertEqual(response.status_code, 302)
        self.assertEqual(response.body, "")
        response_mock.add_header.assert_called_with("Location", location_uri)

    def test_process_not_confirmed(self):
        """
        AuthorizationCodeAuthHandler.process should call SiteAdapter.render_auth_page if the user could not be authenticated
        """
        environ = {"session": "data"}
        response_mock = Mock(spec=Response)
        scopes = ["scopes"]

        client_mock = Mock(spec=Client)

        request_mock = Mock(spec=Request)

        scope_handler_mock = Mock(Scope)
        scope_handler_mock.scopes = scopes

        site_adapter_mock = Mock(spec=AuthorizationCodeGrantSiteAdapter)
        site_adapter_mock.authenticate.side_effect = UserNotAuthenticated
        site_adapter_mock.render_auth_page.return_value = response_mock

        handler = AuthorizationCodeAuthHandler(
            auth_token_store=Mock(), client_authenticator=Mock(),
            scope_handler=scope_handler_mock, site_adapter=site_adapter_mock,
            token_generator=Mock()
        )
        handler.client = client_mock
        response = handler.process(request_mock, response_mock, environ)

        site_adapter_mock.render_auth_page.assert_called_with(request_mock,
                                                              response_mock,
                                                              environ,
                                                              scopes,
                                                              client_mock)
        self.assertEqual(response, response_mock)

    def test_redirect_oauth_error(self):
        error_identifier = "eid"
        redirect_uri = "https://callback"

        expected_redirect = "%s?error=%s" % (redirect_uri, error_identifier)

        error_mock = Mock(spec=OAuthInvalidError)
        error_mock.error = error_identifier

        response_mock = Mock(spec=Response)

        handler = AuthorizationCodeAuthHandler(
            auth_token_store=Mock(), client_authenticator=Mock(),
            scope_handler=Mock(), site_adapter=Mock(), token_generator=Mock())
        handler.client = Client(identifier="abc", secret="xyz",
                                redirect_uris=["https://callback"])
        result = handler.handle_error(error_mock, response_mock)

        response_mock.add_header.assert_called_with("Location",
                                                    expected_redirect)
        response_mock.status_code = 302
        response_mock.body = ""
        self.assertEqual(result, response_mock)


class AuthorizationCodeTokenHandlerTestCase(unittest.TestCase):
    def test_read_validate_params(self):
        client_id = "abc"
        client_secret = "t%gH"
        code = "defg"
        data = {"additional": "data"}
        redirect_uri = "http://callback"
        scopes = ["scope"]
        user_id = 123

        auth_code = Mock(AuthorizationCode)
        auth_code.code = code
        auth_code.data = data
        auth_code.is_expired.return_value = False
        auth_code.redirect_uri = redirect_uri
        auth_code.scopes = scopes
        auth_code.user_id = user_id

        auth_code_store_mock = Mock(spec=AuthCodeStore)
        auth_code_store_mock.fetch_by_code.return_value = auth_code

        client = Client(identifier=client_id, secret=client_secret,
                        redirect_uris=[redirect_uri])

        client_auth_mock = Mock(spec=ClientAuthenticator)
        client_auth_mock.by_identifier_secret.return_value = client

        request_mock = Mock(spec=Request)
        request_mock.post_param.side_effect = [code, redirect_uri]

        handler = AuthorizationCodeTokenHandler(
            access_token_store=Mock(spec=AccessTokenStore),
            auth_token_store=auth_code_store_mock,
            client_authenticator=client_auth_mock,
            token_generator=Mock())

        result = handler.read_validate_params(request_mock)

        request_mock.post_param.assert_has_calls([call("code"),
                                                  call("redirect_uri")])
        auth_code_store_mock.fetch_by_code.assert_called_with(code)
        self.assertEqual(handler.client, client)
        self.assertEqual(handler.code, code)
        self.assertEqual(handler.data, data)
        self.assertEqual(handler.redirect_uri, redirect_uri)
        self.assertEqual(handler.scopes, scopes)
        self.assertEqual(handler.user_id, user_id)
        self.assertTrue(result)

    def test_read_validate_params_missing_code(self):
        client_id = "abc"
        client_secret = "t%gH"
        code = None
        redirect_uri = "http://callback"

        client = Client(identifier=client_id, secret=client_secret,
                        redirect_uris=[redirect_uri])

        client_auth_mock = Mock(spec=ClientAuthenticator)
        client_auth_mock.by_identifier_secret.return_value = client

        request_mock = Mock(spec=Request)
        request_mock.post_param.side_effect = [code, redirect_uri]

        handler = AuthorizationCodeTokenHandler(
            access_token_store=Mock(spec=AccessTokenStore),
            auth_token_store=Mock(spec=AuthCodeStore),
            client_authenticator=client_auth_mock,
            token_generator=Mock())

        with self.assertRaises(OAuthInvalidError) as expected:
            handler.read_validate_params(request_mock)

        error = expected.exception

        self.assertEqual(error.error, "invalid_request")
        self.assertEqual(error.explanation,
                         "Missing required parameter in request")

    def test_read_validate_params_unknown_code(self):
        client_id = "abc"
        client_secret = "t%gH"
        code_expected = "defg"
        code_actual = "xyz"
        redirect_uri = "http://callback"

        auth_code_mock = Mock(AuthorizationCode)
        auth_code_mock.code = code_expected

        auth_code_store_mock = Mock(spec=AuthCodeStore)
        auth_code_store_mock.fetch_by_code.return_value = auth_code_mock

        client = Client(identifier=client_id, secret=client_secret,
                        redirect_uris=[redirect_uri])

        client_auth_mock = Mock(spec=ClientAuthenticator)
        client_auth_mock.by_identifier_secret.return_value = client

        request_mock = Mock(spec=Request)
        request_mock.post_param.side_effect = [code_actual, redirect_uri]

        handler = AuthorizationCodeTokenHandler(
            access_token_store=Mock(spec=AccessTokenStore),
            auth_token_store=auth_code_store_mock,
            client_authenticator=client_auth_mock,
            token_generator=Mock())

        with self.assertRaises(OAuthInvalidError) as expected:
            handler.read_validate_params(request_mock)

        error = expected.exception

        self.assertEqual(error.error, "invalid_grant")
        self.assertEqual(error.explanation,
                         "Invalid code parameter in request")

    def test_read_validate_params_no_auth_code_found(self):
        client_id = "abc"
        client_secret = "t%gH"
        code = "xyz"
        redirect_uri = "http://callback"

        auth_code_store_mock = Mock(spec=AuthCodeStore)
        auth_code_store_mock.fetch_by_code.side_effect = AuthCodeNotFound

        client = Client(identifier=client_id, secret=client_secret,
                        redirect_uris=[redirect_uri])

        client_auth_mock = Mock(spec=ClientAuthenticator)
        client_auth_mock.by_identifier_secret.return_value = client

        request_mock = Mock(spec=Request)
        request_mock.post_param.side_effect = [code, redirect_uri]

        handler = AuthorizationCodeTokenHandler(
            access_token_store=Mock(spec=AccessTokenStore),
            auth_token_store=auth_code_store_mock,
            client_authenticator=client_auth_mock,
            token_generator=Mock())

        with self.assertRaises(OAuthInvalidError) as expected:
            handler.read_validate_params(request_mock)

        error = expected.exception

        self.assertEqual(error.error, "invalid_request")
        self.assertEqual(error.explanation,
                         "Invalid authorization code parameter")

    def test_read_validate_params_wrong_redirect_uri_in_code_data(self):
        client_id = "abc"
        code = "xyz"
        redirect_uri_actual = "http://invalid-callback"
        redirect_uri_expected = "http://callback"

        auth_code_mock = Mock(AuthorizationCode)
        auth_code_mock.code = code
        auth_code_mock.redirect_uri = redirect_uri_actual

        auth_code_store_mock = Mock(spec=AuthCodeStore)
        auth_code_store_mock.fetch_by_code.return_value = auth_code_mock

        client = Client(identifier=client_id, secret="xyz",
                        redirect_uris=[redirect_uri_expected])

        client_auth_mock = Mock(spec=ClientAuthenticator)
        client_auth_mock.by_identifier_secret.return_value = client

        request_mock = Mock(spec=Request)
        request_mock.post_param.side_effect = [code, redirect_uri_expected]

        handler = AuthorizationCodeTokenHandler(
            access_token_store=Mock(spec=AccessTokenStore),
            auth_token_store=auth_code_store_mock,
            client_authenticator=client_auth_mock,
            token_generator=Mock())

        with self.assertRaises(OAuthInvalidError) as expected:
            handler.read_validate_params(request_mock)

        error = expected.exception

        self.assertEqual(error.error, "invalid_request")
        self.assertEqual(error.explanation, "Invalid redirect_uri parameter")

    def test_read_validate_params_token_expired(self):
        client_id = "abc"
        client_secret = "t%gH"
        code = "xyz"
        redirect_uri = "http://callback"

        auth_code_mock = Mock(AuthorizationCode)
        auth_code_mock.code = code
        auth_code_mock.redirect_uri = redirect_uri
        auth_code_mock.is_expired.return_value = True

        auth_code_store_mock = Mock(spec=AuthCodeStore)
        auth_code_store_mock.fetch_by_code.return_value = auth_code_mock

        client = Client(identifier=client_id, secret=client_secret,
                        redirect_uris=[redirect_uri])

        client_auth_mock = Mock(spec=ClientAuthenticator)
        client_auth_mock.by_identifier_secret.return_value = client

        request_mock = Mock(spec=Request)
        request_mock.post_param.side_effect = [code, redirect_uri]

        handler = AuthorizationCodeTokenHandler(
            access_token_store=Mock(spec=AccessTokenStore),
            auth_token_store=auth_code_store_mock,
            client_authenticator=client_auth_mock,
            token_generator=Mock())

        with self.assertRaises(OAuthInvalidError) as expected:
            handler.read_validate_params(request_mock)

        error = expected.exception

        self.assertEqual(error.error, "invalid_grant")
        self.assertEqual(error.explanation, "Authorization code has expired")

    def test_process_no_refresh_token(self):
        client_id = "efg"
        data = {"additional": "data"}
        scopes = ["scope:one", "scope:two"]
        token_data = {"access_token": "abcd", "token_type": "Bearer"}

        response_body = copy(token_data)
        response_body["scope"] = " ".join(scopes)

        access_token_store_mock = Mock(spec=AccessTokenStore)
        auth_code_store_mock = Mock(spec=AuthCodeStore)
        client_auth_mock = Mock(spec=ClientStore)

        token_generator_mock = Mock(spec=TokenGenerator)
        token_generator_mock.create_access_token_data.return_value = token_data

        response_mock = Mock(spec=Response)
        response_mock.body = None
        response_mock.status_code = None

        handler = AuthorizationCodeTokenHandler(
            access_token_store=access_token_store_mock,
            auth_token_store=auth_code_store_mock,
            client_authenticator=client_auth_mock,
            token_generator=token_generator_mock)
        handler.client = Client(identifier=client_id, secret="xyz")
        handler.data = data
        handler.scopes = scopes
        response = handler.process(Mock(spec=Request), response_mock, {})

        self.assertIsNotNone(auth_code_store_mock.delete_code.call_args)
        access_token, = access_token_store_mock.save_token.call_args[0]
        self.assertTrue(isinstance(access_token, AccessToken))
        self.assertEqual(access_token.data, data)
        self.assertEqual(access_token.grant_type, "authorization_code")
        self.assertEqual(response.status_code, 200)
        self.assertEqual(response.body, json.dumps(token_data))
        response_mock.add_header.assert_has_calls([call("Content-Type",
                                                        "application/json"),
                                                   call("Cache-Control",
                                                        "no-store"),
                                                   call("Pragma", "no-cache")])

    @patch("time.time", mock_time)
    def test_process_with_refresh_token(self):
        token_data = {"access_token": "abcd", "token_type": "Bearer",
                      "refresh_token": "wxyz", "expires_in": 600}
        client_id = "efg"
        data = {"additional": "data"}
        scopes = ["scope"]

        access_token_store_mock = Mock(spec=AccessTokenStore)
        auth_code_store_mock = Mock(spec=AuthCodeStore)
        client_auth_mock = Mock(spec=ClientAuthenticator)

        token_generator_mock = Mock(spec=TokenGenerator)
        token_generator_mock.create_access_token_data.return_value = token_data
        token_generator_mock.refresh_expires_in = 10000

        response_mock = Mock(spec=Response)
        response_mock.body = None
        response_mock.status_code = None

        handler = AuthorizationCodeTokenHandler(
            access_token_store=access_token_store_mock,
            auth_token_store=auth_code_store_mock,
            client_authenticator=client_auth_mock,
            token_generator=token_generator_mock)
        handler.client = Client(identifier=client_id, secret="xyz")
        handler.data = data
        handler.scopes = scopes
        response = handler.process(Mock(spec=Request), response_mock, {})

        self.assertIsNotNone(auth_code_store_mock.delete_code.call_args)
        access_token, = access_token_store_mock.save_token.call_args[0]
        self.assertTrue(isinstance(access_token, AccessToken))
        self.assertEqual(access_token.data, data)
        self.assertEqual(access_token.grant_type, "authorization_code")
        self.assertEqual(access_token.expires_at, 1600)
        self.assertEqual(access_token.refresh_token,
                         token_data["refresh_token"])
        self.assertEqual(response.status_code, 200)
        self.assertEqual(response.body, json.dumps(token_data))
        response_mock.add_header.assert_has_calls([call("Content-Type",
                                                        "application/json"),
                                                   call("Cache-Control",
                                                        "no-store"),
                                                   call("Pragma", "no-cache")])

    @patch("time.time", mock_time)
    def test_process_with_unique_access_token(self):
        token_data = {"client_id": "myclient",
                      "grant_type": "authorization_code",
                      "token": "abc123", "data": {}, "expires_at": 1100,
                      "refresh_token": "def456", "scopes": ["foo", "bar"],
                      "user_id": 123}
        token = AccessToken(**token_data)
        expected_response_body = {"access_token": token_data["token"],
                                  "token_type": "Bearer",
                                  "refresh_token": token_data["refresh_token"],
                                  "expires_in": 100, "scope": "foo bar"}

        access_token_store_mock = Mock(spec=AccessTokenStore)
        access_token_store_mock.fetch_existing_token_of_user.return_value = token

        response = Response()

        handler = AuthorizationCodeTokenHandler(
            access_token_store=access_token_store_mock,
            auth_token_store=Mock(spec=AuthCodeStore),
            client_authenticator=Mock(spec=ClientAuthenticator),
            token_generator=Mock())
        handler.client = Client(identifier=token_data["client_id"],
                                secret="xyz")
        handler.data = {}
        handler.scopes = ["foo", "bar"]
        handler.unique_token = True
        handler.user_id = 123

        response = handler.process(request=Mock(spec=Request),
                                   response=response, environ={})
        self.assertDictEqual(json.loads(response.body),
                             expected_response_body)

    @patch("time.time", mock_time)
    def test_process_with_unique_access_token_not_found(self):
        token_data = {"access_token": "abc123", "token_type": "Bearer",
                      "refresh_token": "def456", "expires_in": 1000}
        expected_response_body = copy(token_data)
        expected_response_body["scope"] = "foo bar"

        response = Response()

        access_token_store_mock = Mock(spec=AccessTokenStore)
        access_token_store_mock.fetch_existing_token_of_user. \
            side_effect = AccessTokenNotFound

        token_generator_mock = Mock(spec=TokenGenerator)
        token_generator_mock.refresh_expires_in = 10000
        token_generator_mock.create_access_token_data. \
            return_value = token_data

        handler = AuthorizationCodeTokenHandler(
            access_token_store=access_token_store_mock,
            auth_token_store=Mock(spec=AuthCodeStore),
            client_authenticator=Mock(spec=ClientAuthenticator),
            token_generator=token_generator_mock)
        handler.client = Client(identifier="abc", secret="xyz")
        handler.unique_token = True
        handler.user_id = 123
        handler.scopes = ["foo", "bar"]

        response_result = handler.process(Mock(), response, {})
        self.assertDictEqual(expected_response_body,
                             json.loads(response_result.body))

    def test_process_with_unique_access_token_different_scope(self):
        access_token_data = {"client_id": "myclient",
                             "grant_type": "authorization_code",
                             "token": "xyz890", "data": {}, "expires_at": 1200,
                             "refresh_token": "mno789", "scopes": ["foo", "bar"],
                             "user_id": 123}
        access_token = AccessToken(**access_token_data)
        token_data = {"access_token": "abc123", "token_type": "Bearer",
                      "refresh_token": "def456", "expires_in": 1000}
        expected_response_body = copy(token_data)
        expected_response_body["scope"] = "bar baz"

        response = Response()

        access_token_store_mock = Mock(spec=AccessTokenStore)
        access_token_store_mock.fetch_existing_token_of_user.return_value = access_token

        token_generator_mock = Mock(spec=TokenGenerator)
        token_generator_mock.create_access_token_data.return_value = token_data
        token_generator_mock.refresh_expires_in = 10000

        handler = AuthorizationCodeTokenHandler(
            access_token_store=access_token_store_mock,
            auth_token_store=Mock(spec=AuthCodeStore),
            client_authenticator=Mock(spec=ClientAuthenticator),
            token_generator=token_generator_mock)
        handler.client = Client(identifier=access_token_data["client_id"],
                                secret="xyz")
        handler.data = {}
        handler.unique_token = True
        handler.user_id = 123
        handler.scopes = ["bar", "baz"]

        response_result = handler.process(Mock(), response, {})
        self.assertDictEqual(expected_response_body,
                             json.loads(response_result.body))

    @patch("time.time", mock_time)
    def test_process_with_unique_access_token_expired_token(self):
        access_token_data = {"client_id": "myclient",
                             "grant_type": "authorization_code",
                             "token": "xyz890", "data": {}, "expires_at": 300,
                             "refresh_token": "mno789", "scopes": ["foo", "bar"],
                             "user_id": 123}
        access_token = AccessToken(**access_token_data)
        token_data = {"access_token": "abc123", "token_type": "Bearer",
                      "refresh_token": "def456", "expires_in": 1000}
        expected_response_body = copy(token_data)
        expected_response_body["scope"] = "foo bar"

        response = Response()

        access_token_store_mock = Mock(spec=AccessTokenStore)
        access_token_store_mock.fetch_existing_token_of_user.return_value = access_token

        token_generator_mock = Mock(spec=TokenGenerator)
        token_generator_mock.create_access_token_data.return_value = token_data
        token_generator_mock.refresh_expires_in = 10000

        handler = AuthorizationCodeTokenHandler(
            access_token_store=access_token_store_mock,
            auth_token_store=Mock(spec=AuthCodeStore),
            client_authenticator=Mock(spec=ClientAuthenticator),
            token_generator=token_generator_mock)
        handler.client = Client(identifier=access_token_data["client_id"],
                                secret="xyz")
        handler.data = {}
        handler.unique_token = True
        handler.user_id = 123
        handler.scopes = ["foo", "bar"]

        response_result = handler.process(Mock(), response, {})
        self.assertDictEqual(expected_response_body,
                             json.loads(response_result.body))

    def test_process_with_unique_access_token_no_user_id(self):
        handler = AuthorizationCodeTokenHandler(
            access_token_store=Mock(spec=AccessTokenStore),
            auth_token_store=Mock(spec=AuthCodeStore),
            client_authenticator=Mock(spec=ClientAuthenticator),
            token_generator=Mock(spec=TokenGenerator))
        handler.client = Client(identifier="abc", secret="xyz")
        handler.unique_token = True
        handler.user_id = None

        with self.assertRaises(UserIdentifierMissingError):
            handler.process(Mock(), Mock(), {})


class ImplicitGrantTestCase(unittest.TestCase):
    def test_create_matching_response_type(self):
        path = "/auth"

        request_mock = Mock(spec=Request)
        request_mock.path = path
        request_mock.get_param.return_value = "token"

        token_generator_mock = Mock()

        server_mock = Mock()
        server_mock.authorize_path = path
        server_mock.client_authenticator = Mock()
        server_mock.token_generator = token_generator_mock

        factory = ImplicitGrant(site_adapter=ImplicitGrantSiteAdapter())
        result_class = factory(request_mock, server_mock)

        request_mock.get_param.assert_called_with("response_type")
        self.assertTrue(isinstance(result_class, ImplicitGrantHandler))

    def test_create_not_matching_response_type(self):
        request_mock = Mock(spec=Request)
        request_mock.get_param.return_value = "something"

        server_mock = Mock()

        factory = ImplicitGrant(site_adapter=ImplicitGrantSiteAdapter())
        result_class = factory(request_mock, server_mock)

        request_mock.get_param.assert_called_with("response_type")
        self.assertEqual(result_class, None)


class ImplicitGrantHandlerTestCase(unittest.TestCase):
    def test_process_redirect_with_token(self):
        environ = {"session": "data"}
        redirect_uri = "http://callback"
        scopes = ["scopes"]
        token = "tokencode"
        user_data = ({}, 1)

        access_token_store_mock = Mock(spec=AccessTokenStore)

        client = Client(identifier="abc", secret="xyz",
                        redirect_uris=[redirect_uri])

        request_mock = Mock(spec=Request)

        responseMock = Mock(spec=Response)

        scope_handler_mock = Mock(Scope)
        scope_handler_mock.scopes = scopes
        scope_handler_mock.send_back = False

        site_adapter_mock = Mock(spec=ImplicitGrantSiteAdapter)
        site_adapter_mock.authenticate.return_value = user_data

        token_generator_mock = Mock(spec=["generate"])
        token_generator_mock.generate.return_value = token

        redirect_uri_with_token = "%s#access_token=%s&token_type=bearer" % (redirect_uri, token)

        handler = ImplicitGrantHandler(
            access_token_store=access_token_store_mock,
            client_authenticator=Mock(), scope_handler=scope_handler_mock,
            site_adapter=site_adapter_mock,
            token_generator=token_generator_mock)
        handler.client = client
        result_response = handler.process(request_mock, responseMock, environ)

        site_adapter_mock.authenticate.assert_called_with(request_mock,
                                                          environ, scopes,
                                                          client)

        access_token, = access_token_store_mock.save_token.call_args[0]
        self.assertTrue(isinstance(access_token, AccessToken))
        self.assertEqual(access_token.grant_type, "implicit")

        responseMock.add_header.assert_called_with("Location",
                                                   redirect_uri_with_token)
        self.assertEqual(responseMock.status_code, 302)
        self.assertEqual(responseMock.content, "")
        self.assertEqual(result_response, responseMock)

    def test_process_redirect_with_state(self):
        """
        ImplicitGrantHandler should include the value of the "state" query parameter from request in redirect
        """
        redirect_uri = "http://callback"
        state = "XH%GFI"
        token = "tokencode"
        user_data = ({}, 1)

        expected_redirect_uri = "%s#access_token=%s&token_type=bearer&state=%s" % (redirect_uri, token, quote(state))

        response_mock = Mock(spec=Response)

        scope_handler_mock = Mock(Scope)
        scope_handler_mock.scopes = []
        scope_handler_mock.send_back = False

        site_adapter_mock = Mock(spec=ImplicitGrantSiteAdapter)
        site_adapter_mock.authenticate.return_value = user_data

        token_generator_mock = Mock(spec=["generate"])
        token_generator_mock.generate.return_value = token

        handler = ImplicitGrantHandler(
            access_token_store=Mock(AccessTokenStore),
            client_authenticator=Mock(), scope_handler=scope_handler_mock,
            site_adapter=site_adapter_mock,
            token_generator=token_generator_mock)
        handler.client = Client(identifier="abc", secret="xyz",
                                redirect_uris=[redirect_uri])
        handler.state = state

        result_response = handler.process(request=Mock(spec=Request),
                                          response=response_mock, environ={})

        response_mock.add_header.assert_called_with("Location",
                                                    expected_redirect_uri)
        self.assertEqual(response_mock.status_code, 302)
        self.assertEqual(response_mock.content, "")
        self.assertEqual(result_response, response_mock)

    def test_process_with_scope(self):
        client_id = "abc"
        redirect_uri = "http://callback"
        scopes = ["scope_read", "scope_write"]
        scopes_uri = "%20".join(scopes)
        state = "XHGFI"
        token = "tokencode"

        expected_redirect_uri = "%s#access_token=%s&token_type=bearer&state=%s&scope=%s" % (
            redirect_uri, token, state, scopes_uri)

        response_mock = Mock(spec=Response)

        scope_handler_mock = Mock(Scope)
        scope_handler_mock.scopes = scopes
        scope_handler_mock.send_back = True

        site_adapter_mock = Mock(spec=ImplicitGrantSiteAdapter)
        site_adapter_mock.authenticate.return_value = ({}, 1)

        token_generator_mock = Mock(spec=["generate"])
        token_generator_mock.generate.return_value = token

        handler = ImplicitGrantHandler(
            access_token_store=Mock(AccessTokenStore),
            client_authenticator=Mock(), scope_handler=scope_handler_mock,
            site_adapter=site_adapter_mock,
            token_generator=token_generator_mock)
        handler.client = Client(identifier="abc", secret="xyz",
                                redirect_uris=[redirect_uri])
        handler.state = state

        result_response = handler.process(request=Mock(spec=Request),
                                          response=response_mock, environ={})

        response_mock.add_header.assert_called_with("Location",
                                                    expected_redirect_uri)
        self.assertEqual(response_mock.status_code, 302)
        self.assertEqual(response_mock.content, "")
        self.assertEqual(result_response, response_mock)

    def test_process_unconfirmed(self):
        scopes = ["scopes"]
        environ = {"session": "data"}

        client_mock = Mock(spec=Client)

        request_mock = Mock(spec=Request)

        response_mock = Mock(spec=Response)

        scope_handler_mock = Mock(Scope)
        scope_handler_mock.scopes = scopes

        site_adapter_mock = Mock(spec=ImplicitGrantSiteAdapter)
        site_adapter_mock.authenticate.side_effect = UserNotAuthenticated
        site_adapter_mock.render_auth_page.return_value = response_mock

        handler = ImplicitGrantHandler(
            Mock(spec=AccessTokenStore), client_authenticator=Mock(),
            scope_handler=scope_handler_mock, site_adapter=site_adapter_mock,
            token_generator=Mock()
        )
        handler.client = client_mock
        result_response = handler.process(request_mock, response_mock, environ)

        site_adapter_mock.authenticate.assert_called_with(request_mock,
                                                          environ, scopes,
                                                          client_mock)
        site_adapter_mock.render_auth_page.assert_called_with(request_mock,
                                                              response_mock,
                                                              environ,
                                                              scopes,
                                                              client_mock)
        self.assertEqual(result_response, response_mock)

    def test_process_user_denied_access(self):
        request_mock = Mock(spec=Request)

        responseMock = Mock(spec=Response)

        scope_handler_mock = Mock(spec=Scope)
        scope_handler_mock.scopes = []

        site_adapter_mock = Mock(spec=ImplicitGrantSiteAdapter)
        site_adapter_mock.user_has_denied_access.return_value = True

        handler = ImplicitGrantHandler(
            Mock(spec=AccessTokenStore), client_authenticator=Mock(),
            scope_handler=scope_handler_mock, site_adapter=site_adapter_mock,
            token_generator=Mock()
        )

        with self.assertRaises(OAuthInvalidError) as expected:
            handler.process(request_mock, responseMock, {})

        e = expected.exception

        site_adapter_mock.user_has_denied_access.assert_called_with(
            request_mock
        )
        self.assertEqual(e.error, "access_denied")
        self.assertEqual(e.explanation, "Authorization denied by user")

    def test_redirect_oauth_error(self):
        error_code = "error_code"
        redirect_uri = "https://callback"
        expected_redirect_location = "%s#error=%s" % (redirect_uri, error_code)

        error_mock = Mock(spec=OAuthInvalidError)
        error_mock.error = error_code

        request_mock = Mock(spec=Request)

        response_mock = Mock(spec=Response)

        handler = ImplicitGrantHandler(
            Mock(spec=AccessTokenStore),
            client_authenticator=Mock(), scope_handler=Mock(Scope),
            site_adapter=Mock(), token_generator=Mock())
        handler.client = Client(identifier="abc", secret="xyz",
                                redirect_uris=[redirect_uri])
        altered_response = handler.handle_error(error_mock,
                                                response_mock)

        response_mock.add_header.assert_called_with(
            "Location",
            expected_redirect_location)
        self.assertEqual(altered_response.status_code, 302)
        self.assertEqual(altered_response.body, "")


class ResourceOwnerGrantTestCase(unittest.TestCase):
    def test_call(self):
        request_mock = Mock(Request)
        request_mock.post_param.return_value = "password"

        access_token_store_mock = Mock(AccessTokenStore)
        token_generator_mock = Mock()

        server_mock = Mock(Provider)
        server_mock.access_token_store = access_token_store_mock
        server_mock.client_authenticator = Mock(ClientAuthenticator)
        server_mock.token_generator = token_generator_mock

        factory = ResourceOwnerGrant(
            site_adapter=ResourceOwnerGrantSiteAdapter()
        )

        handler = factory(request_mock, server_mock)

        request_mock.post_param.assert_called_with("grant_type")
        self.assertTrue(isinstance(handler, ResourceOwnerGrantHandler))

    def test_call_no_resource_request(self):
        request_mock = Mock(Request)
        request_mock.post_param.return_value = "other"

        server_mock = Mock(Provider)

        factory = ResourceOwnerGrant(
            site_adapter=ResourceOwnerGrantSiteAdapter()
        )

        handler = factory(request_mock, server_mock)

        request_mock.post_param.assert_called_with("grant_type")
        self.assertEqual(handler, None)


class ResourceOwnerGrantHandlerTestCase(unittest.TestCase):
    def test_process(self):
        access_token = "0aef"
        client = Client(identifier="abc", secret="xyz")
        expected_response_body = {"access_token": access_token,
                                  "token_type": "Bearer"}
        scopes = ["scope"]
        token_data = {"access_token": access_token, "token_type": "Bearer"}
        user = {"id": 123}

        access_token_store_mock = Mock(AccessTokenStore)

        request_mock = Mock(Request)

        response_mock = Mock(Response)

        scope_handler_mock = Mock(Scope)
        scope_handler_mock.scopes = scopes
        scope_handler_mock.send_back = False

        site_adapter_mock = Mock(spec=ResourceOwnerGrantSiteAdapter)
        site_adapter_mock.authenticate.return_value = user

        token_generator_mock = Mock(spec=TokenGenerator)
        token_generator_mock.create_access_token_data.return_value = token_data

        handler = ResourceOwnerGrantHandler(
            access_token_store=access_token_store_mock,
            client_authenticator=Mock(ClientAuthenticator),
            scope_handler=scope_handler_mock,
            site_adapter=site_adapter_mock,
            token_generator=token_generator_mock)
        handler.client = client
        result = handler.process(request_mock, response_mock, {})

        site_adapter_mock.authenticate.assert_called_with(request_mock, {},
                                                          scopes, client)
        token_generator_mock.create_access_token_data.assert_called_with(ResourceOwnerGrant.grant_type)
        access_token, = access_token_store_mock.save_token.call_args[0]
        self.assertTrue(isinstance(access_token, AccessToken))
        self.assertEqual(access_token.grant_type,
                         ResourceOwnerGrant.grant_type)
        response_mock.add_header.assert_has_calls([call("Content-Type",
                                                        "application/json"),
                                                   call("Cache-Control",
                                                        "no-store"),
                                                   call("Pragma", "no-cache")])
        self.assertEqual(result.status_code, 200)
        self.assertEqual(json.loads(result.body), expected_response_body)
        self.assertEqual(result, response_mock)

    @patch("time.time", mock_time)
    def test_process_with_refresh_token(self):
        access_token = "0aef"
        client = Client(identifier="abc", secret="xyz")
        expected_response_body = {"access_token": access_token,
                                  "token_type": "Bearer",
                                  "refresh_token": "wxyz", "expires_in": 600}
        scopes = ["scope"]
        token_data = {"access_token": access_token, "token_type": "Bearer",
                      "refresh_token": "wxyz", "expires_in": 600}
        user = ({"test": "data"}, 123)

        access_token_store_mock = Mock(AccessTokenStore)

        request_mock = Mock(Request)

        response_mock = Mock(Response)

        scope_handler_mock = Mock(Scope)
        scope_handler_mock.scopes = scopes
        scope_handler_mock.send_back = False

        site_adapter_mock = Mock(spec=ResourceOwnerGrantSiteAdapter)
        site_adapter_mock.authenticate.return_value = user

        token_generator_mock = Mock(spec=TokenGenerator)
        token_generator_mock.create_access_token_data.return_value = token_data
        token_generator_mock.refresh_expires_in = 1200

        handler = ResourceOwnerGrantHandler(
            access_token_store=access_token_store_mock,
            client_authenticator=Mock(ClientAuthenticator),
            scope_handler=scope_handler_mock,
            site_adapter=site_adapter_mock,
            token_generator=token_generator_mock)
        handler.client = client
        result = handler.process(request_mock, response_mock, {})

        site_adapter_mock.authenticate.assert_called_with(request_mock, {},
                                                          scopes, client)
        token_generator_mock.create_access_token_data.assert_called_with(ResourceOwnerGrant.grant_type)
        access_token, = access_token_store_mock.save_token.call_args[0]
        self.assertTrue(isinstance(access_token, AccessToken))
        self.assertEqual(access_token.user_id, user[1])
        self.assertEqual(access_token.refresh_token, token_data["refresh_token"])
        self.assertEqual(access_token.expires_at, 1600)
        response_mock.add_header.assert_has_calls([call("Content-Type",
                                                        "application/json"),
                                                   call("Cache-Control",
                                                        "no-store"),
                                                   call("Pragma", "no-cache")])
        self.assertEqual(result.status_code, 200)
        self.assertEqual(json.loads(result.body), expected_response_body)
        self.assertEqual(result, response_mock)

    def test_process_redirect_with_scope(self):
        access_token = "0aef"
        scopes = ["scope_read", "scope_write"]
        expected_response_body = {"access_token": access_token,
                                  "token_type": "Bearer",
                                  "scope": " ".join(scopes)}
        token_data = {"access_token": access_token, "token_type": "Bearer"}

        response_mock = Mock(Response)

        site_adapter_mock = Mock(spec=ResourceOwnerGrantSiteAdapter)
        site_adapter_mock.authenticate.return_value = ({"test": "data"}, 123)

        scope_handler_mock = Mock(Scope)
        scope_handler_mock.scopes = scopes
        scope_handler_mock.send_back = True

        token_generator_mock = Mock(spec=TokenGenerator)
        token_generator_mock.create_access_token_data.return_value = token_data

        handler = ResourceOwnerGrantHandler(
            access_token_store=Mock(AccessTokenStore),
            client_authenticator=Mock(ClientAuthenticator),
            scope_handler=scope_handler_mock,
            site_adapter=site_adapter_mock,
            token_generator=token_generator_mock)
        handler.client = Client(identifier="abc", secret="xyz")
        result = handler.process(Mock(Request), response_mock, {})

        token_generator_mock.create_access_token_data.assert_called_with(ResourceOwnerGrant.grant_type)
        response_mock.add_header.assert_has_calls([call("Content-Type",
                                                        "application/json"),
                                                   call("Cache-Control",
                                                        "no-store"),
                                                   call("Pragma", "no-cache")])
        self.assertEqual(result.status_code, 200)
        self.assertDictEqual(expected_response_body, json.loads(result.body))
        self.assertEqual(result, response_mock)

    def test_process_invalid_user(self):
        response_mock = Mock(Response)

        scope_handler_mock = Mock(Scope)
        scope_handler_mock.scopes = ["scopes"]
        scope_handler_mock.send_back = False

        site_adapter_mock = Mock(spec=ResourceOwnerGrantSiteAdapter)

        site_adapter_mock.authenticate.side_effect = UserNotAuthenticated

        handler = ResourceOwnerGrantHandler(
            access_token_store=Mock(AccessTokenStore),
            client_authenticator=Mock(ClientAuthenticator),
            scope_handler=scope_handler_mock,
            site_adapter=site_adapter_mock,
            token_generator=Mock(TokenGenerator))
        handler.client = Client(identifier="abc", secret="xyz")

        with self.assertRaises(OAuthInvalidError) as expected:
            handler.process(Mock(Request), response_mock, {})

        e = expected.exception

        self.assertEqual(e.error, "invalid_client")
        self.assertEqual(e.explanation,
                         ResourceOwnerGrantHandler.OWNER_NOT_AUTHENTICATED)

    def test_read_validate_params(self):
        password = "johnpw"
        username = "johndoe"

        client = Client(identifier="abcd", secret="xyz")

        client_auth_mock = Mock(ClientAuthenticator)
        client_auth_mock.by_identifier_secret.return_value = client

        request_mock = Mock(Request)
        request_mock.post_param.side_effect = [password, username]

        scope_handler_mock = Mock(Scope)

        handler = ResourceOwnerGrantHandler(
            access_token_store=Mock(AccessTokenStore),
            client_authenticator=client_auth_mock,
            scope_handler=scope_handler_mock,
            site_adapter=Mock(spec=ResourceOwnerGrantSiteAdapter),
            token_generator=Mock())
        result = handler.read_validate_params(request_mock)

        client_auth_mock.by_identifier_secret.assert_called_with(request_mock)
        scope_handler_mock.parse.assert_called_with(request=request_mock,
                                                    source="body")

        self.assertEqual(handler.client, client)
        self.assertEqual(handler.username, username)
        self.assertEqual(handler.password, password)
        self.assertTrue(result)

    def test_handle_error_owner_not_authenticated(self):
        error = OAuthInvalidError(
            error="invalid_client",
            explanation=ResourceOwnerGrantHandler.OWNER_NOT_AUTHENTICATED)

        response = Response()

        handler = ResourceOwnerGrantHandler(
            access_token_store=Mock(AccessTokenStore),
            client_authenticator=Mock(),
            scope_handler=Mock(),
            site_adapter=Mock(),
            token_generator=Mock())

        result = handler.handle_error(error, response)

        self.assertEqual(result.status_code, 401)


class ScopeTestCase(unittest.TestCase):
    def test_parse_scope_scope_present_in_query(self):
        """
        Scope.parse should return a list of requested scopes
        """
        expected_scopes = ["friends_read", "user_read"]

        request_mock = Mock(Request)
        request_mock.get_param.return_value = "friends_read user_read"

        scope = Scope(available=["user_read", "friends_write", "friends_read"])

        scope.parse(request=request_mock, source="query")

        request_mock.get_param.assert_called_with("scope")

        self.assertListEqual(expected_scopes, scope.scopes)
        self.assertFalse(scope.send_back)

    def test_parse_scope_scope_present_in_body(self):
        scope = Scope()

        request_mock = Mock(Request)
        request_mock.post_param.return_value = None

        scope.parse(request=request_mock, source="body")

        request_mock.post_param.assert_called_with("scope")

    def test_parse_scope_default_on_no_scope(self):
        """
        Scope.parse should return a list containing the default value if no scope present in request and default is set
        """
        expected_scopes = ["all"]

        request_mock = Mock(Request)
        request_mock.get_param.return_value = None

        scope = Scope(available=["user_read", "friends_write", "friends_read"],
                      default="all")

        scope.parse(request=request_mock, source="query")

        request_mock.get_param.assert_called_with("scope")

        self.assertListEqual(expected_scopes, scope.scopes)
        self.assertTrue(scope.send_back)

    def test_parse_scope_default_on_no_matching_scopes(self):
        """
        Scope.parse should return a list containing the default value if scope in request does not match and default is set
        """
        expected_scopes = ["all"]

        request_mock = Mock(Request)
        request_mock.get_param.return_value = "user_write"

        scope = Scope(available=["user_read", "friends_write", "friends_read"],
                      default="all")

        scope.parse(request=request_mock, source="query")

        request_mock.get_param.assert_called_with("scope")

        self.assertListEqual(expected_scopes, scope.scopes)
        self.assertTrue(scope.send_back)

    def test_parse_scope_no_value_on_no_scope_no_default(self):
        """
        Scope.parse should return an empty list if no scope is present in request and no default or scapes are defined
        """
        expected_scopes = []

        request_mock = Mock(Request)
        request_mock.get_param.return_value = None

        scope = Scope()

        scope.parse(request=request_mock, source="query")

        request_mock.get_param.assert_called_with("scope")

        self.assertEqual(expected_scopes, scope.scopes)
        self.assertFalse(scope.send_back)

    def test_parse_scope_exception_on_available_scopes_no_scope_given(self):
        """
        Scope.parse should throw an OAuthError if no scope is present in request but scopes are defined
        """
        request_mock = Mock(Request)
        request_mock.get_param.return_value = None

        scope = Scope(available=["user_read", "friends_write", "friends_read"])

        with self.assertRaises(OAuthInvalidError) as expected:
            scope.parse(request_mock, source="query")

        e = expected.exception

        self.assertEqual(e.error, "invalid_scope")

    def test_compare_scopes_equal(self):
        """
        Scope.compare should use the same scopes if new and old scopes do not differ
        """
        scope = Scope(available=["a", "b"])

        scope.scopes = ["a", "b"]

        result = scope.compare(["a", "b"])

        self.assertTrue(result)
        self.assertListEqual(scope.scopes, ["a", "b"])

    def test_compare_valid_scope_subset(self):
        """
        Scope.compare should set a new value for scopes attribute if the new scopes are a subset of the previously issued scopes
        """
        scope = Scope(available=["a", "b", "c"])

        scope.scopes = ["b", "c"]

        result = scope.compare(["a", "b", "c"])

        self.assertTrue(result)
        self.assertListEqual(scope.scopes, ["b", "c"])

    def test_compare_invalid_scope_requested(self):
        """
        Scope.compare should thow an error if a scope is requested that is not contained in the previous scopes.
        """
        scope = Scope(available=["a", "b", "c"])

        scope.scopes = ["b", "c"]

        with self.assertRaises(OAuthInvalidError) as expected:
            scope.compare(["a", "b"])

        e = expected.exception

        self.assertEqual(e.error, "invalid_scope")


class RefreshTokenTestCase(unittest.TestCase):
    def test_call(self):
        """
        RefreshToken should create a new instance of RefreshTokenHandler
        """
        path = "/token"
        expires_in = 600

        access_token_store_mock = Mock()
        client_auth_mock = Mock()
        scope_handler_mock = Mock()
        token_generator_mock = Mock()

        controller_mock = Mock(spec=Provider)
        controller_mock.token_path = path
        controller_mock.access_token_store = access_token_store_mock
        controller_mock.client_authenticator = client_auth_mock
        controller_mock.scope_handler = scope_handler_mock
        controller_mock.token_generator = token_generator_mock
        controller_mock.tokens_expire_in = expires_in

        request_mock = Mock(spec=Request)
        request_mock.path = path
        request_mock.post_param.return_value = "refresh_token"

        grant = RefreshToken(expires_in=0)

        grant_handler = grant(request_mock, controller_mock)

        request_mock.post_param.assert_called_with("grant_type")

        self.assertTrue(isinstance(grant_handler, RefreshTokenHandler))
        self.assertTrue(isinstance(grant_handler.scope_handler, Scope))
        self.assertEqual(access_token_store_mock,
                         grant_handler.access_token_store)
        self.assertEqual(client_auth_mock, grant_handler.client_authenticator)
        self.assertEqual(token_generator_mock, grant_handler.token_generator)

    def test_call_wrong_path(self):
        """
        RefreshToken should return 'None' if path in the request does not equal the token path
        """
        controller_mock = Mock(spec=Provider)
        controller_mock.token_path = "/token"

        request_mock = Mock(spec=Request)
        request_mock.path = "/authorize"

        grant = RefreshToken(expires_in=0)

        grant_handler = grant(request_mock, controller_mock)

        self.assertEqual(grant_handler, None)

    def test_call_other_grant_type(self):
        """
        RefreshToken should return 'None' if another grant type is requested
        """
        path = "/token"

        controller_mock = Mock(spec=Provider)
        controller_mock.token_path = path

        request_mock = Mock(spec=Request)
        request_mock.path = path
        request_mock.get_param.return_value = "authorization_code"

        grant = RefreshToken(expires_in=0)

        grant_handler = grant(request_mock, controller_mock)

        self.assertEqual(grant_handler, None)


class RefreshTokenHandlerTestCase(unittest.TestCase):
    @patch("time.time", mock_time)
    def test_process_no_reissue(self):
        client_id = "testclient"
        data = {"additional": "data"}
        expires_in = 600
        scopes = []
        token = "abcdefg"
        expected_response_body = {"access_token": token,
                                  "expires_in": expires_in,
                                  "token_type": "Bearer"}
        expected_headers = {"Content-Type": "application/json",
                            "Cache-Control": "no-store", "Pragma": "no-cache"}

        access_token_store_mock = Mock(spec=AccessTokenStore)

        response = Response()

        scope_handler_mock = Mock(spec=Scope)
        scope_handler_mock.scopes = scopes

        token_data = {"access_token": token, "expires_in": expires_in, "token_type": "Bearer", "refresh_token": "gafc"}
        token_generator_mock = Mock(spec=TokenGenerator)
        token_generator_mock.create_access_token_data.return_value = token_data
        token_generator_mock.refresh_expires_in = 1200

        handler = RefreshTokenHandler(
            access_token_store=access_token_store_mock,
            client_authenticator=Mock(spec=ClientAuthenticator),
            scope_handler=scope_handler_mock,
            token_generator=token_generator_mock)
        handler.client = Client(identifier=client_id, secret="xyz")
        handler.data = data
        handler.refresh_grant_type = 'test_grant_type'

        result = handler.process(request=Mock(spec=Request),
                                 response=response, environ={})

        access_token, = access_token_store_mock.save_token.call_args[0]
        self.assertEqual(access_token.client_id, client_id)
        self.assertEqual(access_token.grant_type, handler.refresh_grant_type)
        self.assertDictEqual(access_token.data, data)
        self.assertEqual(access_token.token, token)
        self.assertListEqual(access_token.scopes, scopes)
        self.assertEqual(access_token.expires_at, 1600)

        self.assertEqual(result, response)
        self.assertDictContainsSubset(expected_headers, result.headers)
        self.assertEqual(json.dumps(expected_response_body), result.body)

    @patch("time.time", mock_time)
    def test_process_with_reissue(self):
        client_id = "testclient"
        data = {"additional": "data"}
        expires_in = 600
        scopes = []
        token = "abcdefg"
        refresh_token = "cefg"
        expected_response_body = {"access_token": token,
                                  "expires_in": expires_in,
                                  "refresh_token": refresh_token,
                                  "token_type": "Bearer"}
        expected_headers = {"Content-Type": "application/json",
                            "Cache-Control": "no-store", "Pragma": "no-cache"}

        access_token_store_mock = Mock(spec=AccessTokenStore)

        response = Response()

        scope_handler_mock = Mock(spec=Scope)
        scope_handler_mock.scopes = scopes

        token_data = {"access_token": token, "expires_in": expires_in, "token_type": "Bearer",
                      "refresh_token": refresh_token}
        token_generator_mock = Mock(spec=TokenGenerator)
        token_generator_mock.create_access_token_data.return_value = token_data
        token_generator_mock.refresh_expires_in = 1200

        handler = RefreshTokenHandler(
            access_token_store=access_token_store_mock,
            client_authenticator=Mock(spec=ClientAuthenticator),
            scope_handler=scope_handler_mock,
            token_generator=token_generator_mock,
            reissue_refresh_tokens=True)
        handler.client = Client(identifier=client_id, secret="xyz")
        handler.data = data
        handler.refresh_grant_type = 'test_grant_type'

        result = handler.process(request=Mock(spec=Request),
                                 response=response, environ={})

        access_token, = access_token_store_mock.save_token.call_args[0]
        self.assertEqual(access_token.client_id, client_id)
        self.assertEqual(access_token.grant_type, handler.refresh_grant_type)
        self.assertDictEqual(access_token.data, data)
        self.assertEqual(access_token.token, token)
        self.assertListEqual(access_token.scopes, scopes)
        self.assertEqual(access_token.expires_at, 1600)

        self.assertEqual(result, response)
        self.assertDictContainsSubset(expected_headers, result.headers)
        self.assertDictEqual(expected_response_body, json.loads(result.body))

    @patch("time.time", mock_time)
    def test_read_validate_params(self):
        client_id = "client"
        client_secret = "secret"
        data = {"additional": "data"}
        original_token = "sd3f3j"
        refresh_token = "s74jf"
        scopes = []

        access_token = AccessToken(client_id=client_id, token=original_token,
                                   grant_type="test_grant_type",
                                   data=data, expires_at=1234, scopes=scopes,
                                   refresh_expires_at=0)

        access_token_store_mock = Mock(AccessTokenStore)
        access_token_store_mock.fetch_by_refresh_token.return_value = access_token

        client = Client(identifier=client_id, secret=client_secret,
                        redirect_uris=[])

        client_auth_mock = Mock(spec=ClientAuthenticator)
        client_auth_mock.by_identifier_secret.return_value = client

        request_mock = Mock(spec=Request)
        request_mock.post_param.side_effect = [refresh_token]

        token_generator_mock = Mock(expires_in={'test_grant_type': 600})
        token_generator_mock.refresh_expires_in = 0

        scope_handler_mock = Mock(spec=Scope)

        handler = RefreshTokenHandler(
            access_token_store=access_token_store_mock,
            client_authenticator=client_auth_mock,
            scope_handler=scope_handler_mock,
            token_generator=token_generator_mock)

        handler.read_validate_params(request=request_mock)

        request_mock.post_param.assert_called_with("refresh_token")
        access_token_store_mock.fetch_by_refresh_token.assert_called_with(refresh_token)
        client_auth_mock.by_identifier_secret.assert_called_with(request_mock)
        scope_handler_mock.parse.assert_called_with(request_mock, "body")
        scope_handler_mock.compare.assert_called_with(scopes)

        self.assertEqual(handler.client, client)
        self.assertEqual(handler.data, data)
        self.assertEqual(handler.refresh_token, refresh_token)

    def test_read_validate_params_no_refresh_token(self):
        request_mock = Mock(spec=Request)
        request_mock.post_param.return_value = None

        handler = RefreshTokenHandler(access_token_store=Mock(),
                                      client_authenticator=Mock(),
                                      scope_handler=Mock(),
                                      token_generator=Mock(expires_in=600))

        with self.assertRaises(OAuthInvalidError) as expected:
            handler.read_validate_params(request_mock)

        e = expected.exception

        self.assertEqual(e.error, "invalid_request")
        self.assertEqual(e.explanation,
                         "Missing refresh_token in request body")

    def test_read_validate_params_invalid_refresh_token(self):
        client_id = "abc"
        secret = "xyz"

        access_token_store_mock = Mock(spec=AccessTokenStore)
        access_token_store_mock.fetch_by_refresh_token.side_effect = AccessTokenNotFound

        request_mock = Mock(spec=Request)
        request_mock.post_param.side_effect = [client_id, secret, "uuu"]

        client = Client(identifier=client_id, secret=secret, redirect_uris=[])

        client_auth_mock = Mock(spec=ClientAuthenticator)
        client_auth_mock.by_identifier_secret.return_value = client

        handler = RefreshTokenHandler(access_token_store=access_token_store_mock,
                                      client_authenticator=client_auth_mock,
                                      scope_handler=Mock(),
                                      token_generator=Mock(expires_in=600))

        with self.assertRaises(OAuthInvalidError) as expected:
            handler.read_validate_params(request_mock)

        e = expected.exception

        self.assertEqual(e.error, "invalid_request")
        self.assertEqual(e.explanation, "Invalid refresh token")

    @patch("time.time", mock_time)
    def test_read_validate_params_expired_refresh_token(self):
        client_id = "abc"
        secret = "xyz"

        access_token_mock = Mock(spec=AccessToken)
        access_token_mock.grant_type = 'test_grant_type'
        access_token_mock.refresh_expires_at = 900

        access_token_store_mock = Mock(spec=AccessTokenStore)
        access_token_store_mock.fetch_by_refresh_token.return_value = access_token_mock

        request_mock = Mock(spec=Request)
        request_mock.post_param.side_effect = [client_id, secret, "uuu"]

        client = Client(identifier=client_id, secret=secret, redirect_uris=[])

        client_auth_mock = Mock(spec=ClientAuthenticator)
        client_auth_mock.by_identifier_secret.return_value = client

        handler = RefreshTokenHandler(
            access_token_store=access_token_store_mock,
            client_authenticator=client_auth_mock,
            scope_handler=Mock(),
            token_generator=Mock(expires_in={'test_grant_type': 600}))

        with self.assertRaises(OAuthInvalidError) as expected:
            handler.read_validate_params(request_mock)

        e = expected.exception

        self.assertEqual(e.error, "invalid_request")
        self.assertEqual(e.explanation, "Invalid refresh token")


class ClientCredentialsGrantTestCase(unittest.TestCase):
    def test_call(self):
        token_path = "token"

        request_mock = Mock(spec=Request)
        request_mock.path = token_path
        request_mock.post_param.return_value = "client_credentials"

        access_token_store_mock = Mock()
        client_auth_mock = Mock()
        token_generator_mock = Mock()

        scope_handler_mock = Mock(spec=Scope)

        server_mock = Mock()
        server_mock.token_path = token_path
        server_mock.access_token_store = access_token_store_mock
        server_mock.client_authenticator = client_auth_mock
        server_mock.scope_handler = scope_handler_mock
        server_mock.token_generator = token_generator_mock

        grant = ClientCredentialsGrant()
        handler = grant(request_mock, server_mock)

        request_mock.post_param.assert_called_with("grant_type")
        self.assertTrue(isinstance(handler, ClientCredentialsHandler))
        self.assertEqual(handler.access_token_store, access_token_store_mock)
        self.assertEqual(handler.client_authenticator, client_auth_mock)
        self.assertTrue(isinstance(handler.scope_handler, Scope))
        self.assertEqual(handler.token_generator, token_generator_mock)

    def test_call_wrong_request_path(self):
        request_mock = Mock(spec=Request)
        request_mock.path = "authorize"

        server_mock = Mock()
        server_mock.token_path = "token"

        grant = ClientCredentialsGrant()
        handler = grant(request_mock, server_mock)

        self.assertEqual(handler, None)

    def test_call_other_grant_type(self):
        token_path = "token"

        request_mock = Mock(spec=Request)
        request_mock.path = token_path
        request_mock.post_param.return_value = "other_grant"

        server_mock = Mock()
        server_mock.token_path = token_path

        grant = ClientCredentialsGrant()
        handler = grant(request_mock, server_mock)

        self.assertEqual(handler, None)


class ClientCredentialsHandlerTestCase(unittest.TestCase):
    def test_process(self):
        client_id = "abc"
        token = "abcd"

        expected_response_body = {"access_token": token,
                                  "token_type": "Bearer"}

        access_token_store_mock = Mock(spec=AccessTokenStore)

        response_mock = Mock(spec=Response)

        scope_handler_mock = Mock(spec=Scope)
        scope_handler_mock.send_back = False
        scope_handler_mock.scopes = []

        token_generator_mock = Mock(spec=TokenGenerator)
        token_generator_mock.generate.return_value = token
        token_generator_mock.expires_in = {}
        handler = ClientCredentialsHandler(
            access_token_store=access_token_store_mock,
            client_authenticator=Mock(),
            scope_handler=scope_handler_mock,
            token_generator=token_generator_mock)
        handler.client = Client(identifier=client_id, secret="xyz")
        result_response = handler.process(request=Mock(),
                                          response=response_mock, environ={})

        self.assertDictEqual(json.loads(result_response.body),
                             expected_response_body)

    @patch("time.time", mock_time)
    def test_process_with_refresh_token(self):
        client_id = "abc"
        expires_in = 600
        token = "abcd"
        scopes = ["foo", "bar"]

        expected_response_body = {"access_token": token,
                                  "expires_in": expires_in,
                                  "token_type": "Bearer",
                                  "scope": " ".join(scopes)}

        access_token_store_mock = Mock(spec=AccessTokenStore)

        response_mock = Mock(spec=Response)

        scope_handler_mock = Mock(spec=Scope)
        scope_handler_mock.send_back = True
        scope_handler_mock.scopes = scopes

        token_generator_mock = Mock(spec=TokenGenerator)
        token_generator_mock.generate.return_value = token
        token_generator_mock.expires_in = {ClientCredentialsGrant.grant_type: expires_in}

        handler = ClientCredentialsHandler(
            access_token_store=access_token_store_mock,
            client_authenticator=Mock(),
            scope_handler=scope_handler_mock,
            token_generator=token_generator_mock)
        handler.client = Client(identifier=client_id, secret="xyz")
        result_response = handler.process(request=Mock(),
                                          response=response_mock, environ={})

        access_token, = access_token_store_mock.save_token.call_args[0]
        self.assertTrue(isinstance(access_token, AccessToken))
        self.assertEqual(access_token.client_id, client_id)
        self.assertEqual(access_token.grant_type, "client_credentials")
        self.assertEqual(access_token.token, token)
        self.assertEqual(access_token.data, {})
        self.assertEqual(access_token.expires_at, expires_in + 1000)
        self.assertEqual(access_token.refresh_token, None)
        self.assertEqual(access_token.scopes, scopes)

        response_mock.add_header.assert_has_calls([call("Content-Type",
                                                        "application/json"),
                                                   call("Cache-Control",
                                                        "no-store"),
                                                   call("Pragma", "no-cache")])
        self.assertDictEqual(json.loads(result_response.body),
                             expected_response_body)

    def test_read_validate_params(self):
        client_id = "abc"
        client_secret = "xyz"

        client_auth_mock = Mock(spec=ClientAuthenticator)
        client_auth_mock.by_identifier_secret.return_value = Client(
            identifier=client_id,
            secret=client_secret,
            redirect_uris=[])

        scope_handler_mock = Mock(spec=Scope)

        request_mock = Mock(spec=Request)

        handler = ClientCredentialsHandler(
            access_token_store=Mock(),
            client_authenticator=client_auth_mock,
            scope_handler=scope_handler_mock,
            token_generator=Mock())
        handler.read_validate_params(request_mock)

        client_auth_mock.by_identifier_secret.assert_called_with(request_mock)
        scope_handler_mock.parse.assert_called_with(request=request_mock,
                                                    source="body")


if __name__ == "__main__":
    unittest.main()



import re
from oauth2.test import unittest
from oauth2.tokengenerator import URandomTokenGenerator, Uuid4

class URandomTokenGeneratorTestCase(unittest.TestCase):
    def test_generate(self):
        length = 20
        
        generator = URandomTokenGenerator(length=length)
        
        result = generator.generate()
        
        self.assertTrue(isinstance(result, str))
        self.assertEqual(len(result), length)

class Uuid4TestCase(unittest.TestCase):
    def setUp(self):
        self.uuid_regex = r"^[a-z0-9]{8}\-[a-z0-9]{4}\-[a-z0-9]{4}\-[a-z0-9]{4}-[a-z0-9]{12}$"
    
    def test_create_access_token_data_no_expiration(self):
        generator = Uuid4()
        
        result = generator.create_access_token_data('test_grant_type')
        
        self.assertRegexpMatches(result["access_token"], self.uuid_regex)
        self.assertEqual(result["token_type"], "Bearer")
    
    def test_create_access_token_data_with_expiration(self):
        generator = Uuid4()
        generator.expires_in = {'test_grant_type':600}
        
        result = generator.create_access_token_data('test_grant_type')
        
        self.assertRegexpMatches(result["access_token"], self.uuid_regex)
        self.assertEqual(result["token_type"], "Bearer")
        self.assertRegexpMatches(result["refresh_token"], self.uuid_regex)
        self.assertEqual(result["expires_in"], 600)
    
    def test_generate(self):
        generator = Uuid4()
        
        result = generator.generate()
        
        regex = re.compile(self.uuid_regex)
        
        match = regex.match(result)
        
        self.assertEqual(result, match.group())

if __name__ == "__main__":
    unittest.main()



# -*- coding: utf-8 -*-
import redis
import json

from oauth2.datatype import AccessToken, AuthorizationCode, Client
from oauth2.error import AccessTokenNotFound, AuthCodeNotFound, \
    ClientNotFoundError
from oauth2.store import AccessTokenStore, AuthCodeStore, ClientStore


class RedisStore(object):
    """
    Uses redis to store access tokens and auth tokens.

    This Store supports ``redis``. Arguments are passed to the
    underlying client implementation.

    Initialization::

        import redisdb

        token_store = TokenStore(host="127.0.0.1",
            port=6379,
            db=0
        )

    """
    def __init__(self, rs=None, prefix="oauth2", *args, **kwargs):
        self.prefix = prefix

        if rs is not None:
            self.rs = rs
        else:
            self.rs = redis.StrictRedis(*args, **kwargs)

    def delete(self, name):
        cache_key = self._generate_cache_key(name)

        self.rs.delete(cache_key)

    def write(self, name, data):
        cache_key = self._generate_cache_key(name)

        self.rs.set(cache_key, json.dumps(data))

    def read(self, name):
        cache_key = self._generate_cache_key(name)

        data = self.rs.get(cache_key)

        if data is None:
            return None

        return json.loads(data.decode("utf-8"))

    def _generate_cache_key(self, identifier):
        return self.prefix + "_" + identifier


class TokenStore(AccessTokenStore, AuthCodeStore, RedisStore):
    def fetch_by_code(self, code):
        """
        Returns data belonging to an authorization code from redis or
        ``None`` if no data was found.

        See :class:`oauth2.store.AuthCodeStore`.

        """
        code_data = self.read(code)

        if code_data is None:
            raise AuthCodeNotFound

        return AuthorizationCode(**code_data)

    def save_code(self, authorization_code):
        """
        Stores the data belonging to an authorization code token in redis.

        See :class:`oauth2.store.AuthCodeStore`.

        """
        self.write(authorization_code.code,
                   {"client_id": authorization_code.client_id,
                    "code": authorization_code.code,
                    "expires_at": authorization_code.expires_at,
                    "redirect_uri": authorization_code.redirect_uri,
                    "scopes": authorization_code.scopes,
                    "data": authorization_code.data,
                    "user_id": authorization_code.user_id})

    def delete_code(self, code):
        """
        Deletes an authorization code after use
        :param code: The authorization code.
        """
        self.delete(code)

    def save_token(self, access_token):
        """
        Stores the access token and additional data in redis.

        See :class:`oauth2.store.AccessTokenStore`.

        """
        self.write(access_token.token, access_token.__dict__)

        unique_token_key = self._unique_token_key(access_token.client_id,
                                                  access_token.grant_type,
                                                  access_token.user_id)
        self.write(unique_token_key, access_token.__dict__)

        if access_token.refresh_token is not None:
            self.write(access_token.refresh_token, access_token.__dict__)

    def delete_refresh_token(self, refresh_token):
        """
        Deletes a refresh token after use
        :param refresh_token: The refresh token to delete.
        """
        access_token = self.fetch_by_refresh_token(refresh_token)

        self.delete(access_token.token)

    def fetch_by_refresh_token(self, refresh_token):
        token_data = self.read(refresh_token)

        if token_data is None:
            raise AccessTokenNotFound

        return AccessToken(**token_data)

    def fetch_existing_token_of_user(self, client_id, grant_type, user_id):
        unique_token_key = self._unique_token_key(client_id=client_id,
                                                  grant_type=grant_type,
                                                  user_id=user_id)
        token_data = self.read(unique_token_key)

        if token_data is None:
            raise AccessTokenNotFound

        return AccessToken(**token_data)

    def _unique_token_key(self, client_id, grant_type, user_id):
        return "{0}_{1}_{2}".format(client_id, grant_type, user_id)


class ClientStore(ClientStore, RedisStore):
    def add_client(self, client_id, client_secret, redirect_uris,
                   authorized_grants=None, authorized_response_types=None):
        """
        Add a client app.

        :param client_id: Identifier of the client app.
        :param client_secret: Secret the client app uses for authentication
                              against the OAuth 2.0 provider.
        :param redirect_uris: A ``list`` of URIs to redirect to.

        """
        self.write(client_id,
                   {"identifier": client_id,
                    "secret": client_secret,
                    "redirect_uris": redirect_uris,
                    "authorized_grants": authorized_grants,
                    "authorized_response_types": authorized_response_types})

        return True

    def fetch_by_client_id(self, client_id):
        client_data = self.read(client_id)

        if client_data is None:
            raise ClientNotFoundError

        return Client(identifier=client_data["identifier"],
                      secret=client_data["secret"],
                      redirect_uris=client_data["redirect_uris"],
                      authorized_grants=client_data["authorized_grants"],
                      authorized_response_types=client_data["authorized_response_types"])



# -*- coding: utf-8 -*-
"""
Definitions of types used by grants.
"""

import time
from oauth2.error import RedirectUriUnknown


class AccessToken(object):
    """
    An access token and associated data.
    """
    def __init__(self, client_id, grant_type, token, data={}, expires_at=None,
                 refresh_token=None, refresh_expires_at=None, scopes=[],
                 user_id=None):
        self.client_id = client_id
        self.grant_type = grant_type
        self.token = token
        self.data = data
        self.expires_at = expires_at
        self.refresh_token = refresh_token
        self.refresh_expires_at = refresh_expires_at
        self.scopes = scopes
        self.user_id = user_id

    @property
    def expires_in(self):
        """
        Returns the time until the token expires.

        :return: The remaining time until expiration in seconds or 0 if the
                 token has expired.
        """
        time_left = self.expires_at - int(time.time())

        if time_left > 0:
            return time_left
        return 0

    def is_expired(self):
        """
        Determines if the token has expired.

        :return: `True` if the token has expired. Otherwise `False`.
        """
        if self.expires_at is None:
            return False

        if self.expires_in > 0:
            return False

        return True


class AuthorizationCode(object):
    """
    Holds an authorization code and additional information.
    """
    def __init__(self, client_id, code, expires_at, redirect_uri, scopes,
                 data=None, user_id=None):
        self.client_id = client_id
        self.code = code
        self.data = data
        self.expires_at = expires_at
        self.redirect_uri = redirect_uri
        self.scopes = scopes
        self.user_id = user_id

    def is_expired(self):
        if self.expires_at < int(time.time()):
            return True
        return False


class Client(object):
    """
    Representation of a client application.
    """
    def __init__(self, identifier, secret, authorized_grants=None,
                 authorized_response_types=None, redirect_uris=None):
        """
        :param identifier: The unique identifier of a client.
        :param secret: The secret the clients uses to authenticate.
        :param authorized_grants: A list of grants under which the client can
                                  request tokens.
                                  All grants are allowed if this value is set
                                  to `None` (default).
        :param authorized_response_types: A list of response types of which
                                          the client can request tokens.
                                          All response types are allowed if
                                          this value is set to `None`
                                          (default).
        :redirect_uris: A list of redirect uris this client can use.
        """
        self.authorized_grants = authorized_grants
        self.authorized_response_types = authorized_response_types
        self.identifier = identifier
        self.secret = secret

        if redirect_uris is None:
            self.redirect_uris = []
        else:
            self.redirect_uris = redirect_uris

        self._redirect_uri = None

    @property
    def redirect_uri(self):
        if self._redirect_uri is None:
            # redirect_uri is an optional param.
            # If not supplied, we use the first entry stored in db as default.
            return self.redirect_uris[0]
        return self._redirect_uri

    @redirect_uri.setter
    def redirect_uri(self, value):
        if value not in self.redirect_uris:
            raise RedirectUriUnknown
        self._redirect_uri = value

    def grant_type_supported(self, grant_type):
        """
        Checks if the Client is authorized receive tokens for the given grant.

        :param grant_type: The type of the grant.

        :return: Boolean
        """
        if self.authorized_grants is None:
            return True

        return grant_type in self.authorized_grants

    def response_type_supported(self, response_type):
        """
        Checks if the client is allowed to receive tokens for the given
        response type.

        :param response_type: The response type.

        :return: Boolean
        """
        if self.authorized_response_types is None:
            return True

        return response_type in self.authorized_response_types



"""
Grants are the heart of OAuth 2.0. Each Grant defines one way for a client to
retrieve an authorization. They are defined in
`Section 4 <http://tools.ietf.org/html/rfc6749#section-4>`_ of the OAuth 2.0
spec.

OAuth 2.0 comes in two flavours of how an access token is issued:
two-legged and three-legged auth. To avoid confusion they are explained in
short here.

Three-legged OAuth
------------------
The "three" symbolizes the parties that are involved:

* The client that wants to access a resource on behalf of the user.
* The user who grants access to her resources.
* The server that issues the access token if the user allows it.

Two-legged OAuth
----------------
The two-legged OAuth process differs from the three-legged process by one
missing participant. The user cannot allow or deny access.

So there are two remaining parties:

* The client that wants to access a resource.
* The server that issues the access.

"""
from oauth2.error import OAuthInvalidError, UserNotAuthenticated, \
    AccessTokenNotFound, UserIdentifierMissingError, RedirectUriUnknown, \
    AuthCodeNotFound, InvalidSiteAdapter
from oauth2.compatibility import urlencode, quote
import json
import time
from oauth2.datatype import AuthorizationCode, AccessToken
from oauth2.web import Response, AuthorizationCodeGrantSiteAdapter, \
    ImplicitGrantSiteAdapter, ResourceOwnerGrantSiteAdapter


def encode_scopes(scopes, use_quote=False):
    """
    Creates a string out of a list of scopes.

    :param scopes: A list of scopes
    :param use_quote: Boolean flag indicating whether the string should be quoted
    :return: Scopes as a string
    """
    scopes_as_string = Scope.separator.join(scopes)

    if use_quote:
        return quote(scopes_as_string)
    return scopes_as_string


def json_error_response(error, response, status_code=400):
    """
    Formats an error as a response containing a JSON body.
    """
    msg = {"error": error.error, "error_description": error.explanation}

    response.status_code = status_code
    response.add_header("Content-Type", "application/json")
    response.body = json.dumps(msg)

    return response


def json_success_response(data, response):
    """
    Formats the response of a successful token request as JSON.

    Also adds default headers and status code.
    """
    response.body = json.dumps(data)
    response.status_code = 200

    response.add_header("Content-Type", "application/json")
    response.add_header("Cache-Control", "no-store")
    response.add_header("Pragma", "no-cache")


class ResponseTypeGrant(object):
    def error_response(self, response):
        pass


class Scope(object):
    """
    Handling of the "scope" parameter in a request.

    If ``available`` and ``default`` are both ``None``, the "scope" parameter
    is ignored (the default).

    :param available: A list of strings each defining one supported scope.
    :param default: Value to fall back to in case no scope is present in a
                    request.
    """

    separator = " "

    def __init__(self, available=None, default=None):
        self.scopes = []
        self.send_back = False

        if isinstance(available, list):
            self.available_scopes = available
        else:
            self.available_scopes = []

        self.default = default

    def compare(self, previous_scopes):
        """
        Compares the scopes read from request with previously issued scopes.

        :param previous_scopes: A list of scopes.
        :return: ``True``
        """
        for scope in self.scopes:
            if scope not in previous_scopes:
                raise OAuthInvalidError(
                    error="invalid_scope",
                    explanation="Invalid scope parameter in request")

        return True

    def parse(self, request, source):
        """
        Parses scope value in given request.

        Expects the value of the "scope" parameter in request to be a string
        where each requested scope is separated by a white space::

            # One scope requested
            "profile_read"

            # Multiple scopes
            "profile_read profile_write"

        :param request: An instance of :class:`oauth2.web.Request`.
        :param source: Where to read the scope from. Pass "body" in case of a
                       application/x-www-form-urlencoded body and "query" in
                       case the scope is supplied as a query parameter in the
                       URL of a request.
        """
        if source == "body":
            req_scope = request.post_param("scope")
        elif source == "query":
            req_scope = request.get_param("scope")
        else:
            raise ValueError("Unknown scope source '" + source + "'")

        if req_scope is None:
            if self.default is not None:
                self.scopes = [self.default]
                self.send_back = True
                return
            elif len(self.available_scopes) != 0:
                raise OAuthInvalidError(
                    error="invalid_scope",
                    explanation="Missing scope parameter in request")
            else:
                return

        req_scopes = req_scope.split(self.separator)

        self.scopes = [scope for scope in req_scopes
                       if scope in self.available_scopes]

        if len(self.scopes) == 0 and self.default is not None:
            self.scopes = [self.default]
            self.send_back = True


class ScopeGrant(object):
    """
    Handling of scopes in the OAuth 2.0 flow.

    Inherited by all grants that need to support scopes.

    :param default_scope: The scope identifier that is returned by default.
                          (optional)
    :param scopes:        A list of strings identifying the scopes that the
                          grant supports.
    :param scope_class: The class that does the actual handling in a request.
                        Default: :class:`oauth2.grant.Scope`.
    """

    def __init__(self, default_scope=None, scopes=None, scope_class=Scope,
                 **kwargs):
        self.default_scope = default_scope
        self.scopes = scopes
        self.scope_class = scope_class

        super(ScopeGrant, self).__init__(**kwargs)

    def _create_scope_handler(self):
        return self.scope_class(available=self.scopes,
                                default=self.default_scope)


class GrantHandler(object):
    """
    Base class every oauth2 handler can extend.
    """

    def process(self, request, response, environ):
        """
        Handles the logic of how a user gets an access token.

        This includes steps like calling the implementation of a `SiteAdapter`
        if the user is authorized or generating a new access token.

        This method uses data read in `read_validate_params`.
        """
        raise NotImplementedError

    def read_validate_params(self, request):
        """
        Reads and validates the incoming data.
        """
        raise NotImplementedError

    def handle_error(self, error, response):
        """
        Takes all the actions necessary to return an error response in the
        format defined for a specific grant handler.
        """
        raise NotImplementedError


class GrantHandlerFactory(object):
    """
    Base class every handler factory can extend.

    This class defines the basic interface of each Grant.
    """

    def __call__(self, request, server):
        raise NotImplementedError


class AuthRequestMixin(object):
    """
    Generalization of reading and validating an incoming request used by
    `oauth2.grant.AuthorizationCodeAuthHandler` and
    `oauth2.grant.ImplicitGrantHandler`.
    """

    def __init__(self, client_authenticator, scope_handler, token_generator,
                 **kwargs):
        self.client = None
        self.state = None

        self.client_authenticator = client_authenticator
        self.scope_handler = scope_handler
        self.token_generator = token_generator

        super(AuthRequestMixin, self).__init__(**kwargs)

    def read_validate_params(self, request):
        """
        Reads and validates data in an incoming request as required by
        the Authorization Request of the Authorization Code Grant and the
        Implicit Grant.
        """
        self.client = self.client_authenticator.by_identifier(request)

        response_type = request.get_param("response_type")

        if self.client.response_type_supported(response_type) is False:
            raise OAuthInvalidError(error="unauthorized_client")

        self.state = request.get_param("state")

        self.scope_handler.parse(request, "query")

        return True


class AuthorizeMixin(object):
    """
    Used by all grants that involve user interaction.
    """

    def __init__(self, site_adapter, **kwargs):
        self.site_adapter = site_adapter

        super(AuthorizeMixin, self).__init__(**kwargs)

    def authorize(self, request, response, environ, scopes):
        """
        Controls all steps to authorize a request by a user.

        :param request: The incoming :class:`oauth2.web.Request`
        :param response: The :class:`oauth2.web.Response` that will be
                         returned eventually
        :param environ: The environment variables of this request
        :param scopes: The scopes requested by an application
        :return: A tuple containing (`dict`, user_id) or the response.

        """
        if self.site_adapter.user_has_denied_access(request) is True:
            raise OAuthInvalidError(error="access_denied",
                                    explanation="Authorization denied by user")

        try:
            result = self.site_adapter.authenticate(request, environ, scopes,
                                                    self.client)

            return self.sanitize_return_value(result)
        except UserNotAuthenticated:
            return self.site_adapter.render_auth_page(request, response,
                                                      environ, scopes,
                                                      self.client)

    @staticmethod
    def sanitize_return_value(value):
        if isinstance(value, tuple) and len(value) is 2:
            return value

        return value, None


class AccessTokenMixin(object):
    """
    Used by grants that handle refresh token and unique token.
    """

    def __init__(self, access_token_store, token_generator,
                 unique_token=False, **kwargs):
        self.access_token_store = access_token_store
        self.token_generator = token_generator

        self.unique_token = unique_token

        super(AccessTokenMixin, self).__init__(**kwargs)

    def create_token(self, client_id, data, grant_type, scopes, user_id):
        if self.unique_token:
            if user_id is None:
                raise UserIdentifierMissingError

            try:
                access_token = self.access_token_store. \
                    fetch_existing_token_of_user(
                    client_id,
                    grant_type,
                    user_id)

                if (access_token.scopes == scopes
                    and access_token.is_expired() is False):
                    token_data = {"access_token": access_token.token,
                                  "token_type": "Bearer"}

                    if access_token.refresh_token is not None:
                        token_data["refresh_token"] = access_token.refresh_token
                        token_data["expires_in"] = access_token.expires_in

                    return token_data
            except AccessTokenNotFound:
                pass

        token_data = self.token_generator.create_access_token_data(grant_type)

        access_token = AccessToken(client_id=client_id, data=data,
                                   grant_type=grant_type,
                                   token=token_data["access_token"],
                                   scopes=scopes,
                                   user_id=user_id)

        if "refresh_token" in token_data:
            expires_at = int(time.time()) + token_data["expires_in"]
            access_token.expires_at = expires_at
            access_token.refresh_token = token_data["refresh_token"]
            refresh_expires_in = self.token_generator.refresh_expires_in
            refresh_expires_at = int(time.time()) + refresh_expires_in
            access_token.refresh_expires_at = refresh_expires_at

        self.access_token_store.save_token(access_token)

        return token_data


class SiteAdapterMixin(object):
    """
    Mixed in by Grant classes that require a site adapter.

    A concrete class must set the class attribute ``site_adapter_class`` that
    contains a reference to the site adapter class that this class expects.
    """
    site_adapter_class = None

    def __init__(self, site_adapter, **kwargs):
        if isinstance(site_adapter, self.site_adapter_class) is False:
            raise InvalidSiteAdapter(
                "Site adapter must inherit from class '{0}'"
                    .format(self.site_adapter_class.__name__)
            )

        self.site_adapter = site_adapter

        super(SiteAdapterMixin, self).__init__(**kwargs)


class AuthorizationCodeAuthHandler(AuthorizeMixin, AuthRequestMixin,
                                   GrantHandler):
    """
    Implementation of the first step of the Authorization Code Grant
    (three-legged).
    """

    token_expiration = 600

    def __init__(self, auth_token_store, **kwargs):
        self.auth_code_store = auth_token_store

        super(AuthorizationCodeAuthHandler, self).__init__(**kwargs)

    def process(self, request, response, environ):
        """
        Generates a new authorization token.

        A form to authorize the access of the application can be displayed with
        the help of `oauth2.web.SiteAdapter`.
        """
        data = self.authorize(request, response, environ,
                              self.scope_handler.scopes)

        if isinstance(data, Response):
            return data

        code = self.token_generator.generate()
        expires = int(time.time()) + self.token_expiration

        auth_code = AuthorizationCode(client_id=self.client.identifier,
                                      code=code, expires_at=expires,
                                      redirect_uri=self.client.redirect_uri,
                                      scopes=self.scope_handler.scopes,
                                      data=data[0], user_id=data[1])

        self.auth_code_store.save_code(auth_code)

        response.add_header("Location", self._generate_location(code))
        response.body = ""
        response.status_code = 302

        return response

    def handle_error(self, error, response):
        """
        Redirects the client in case an error in the auth process occurred.
        """
        query_params = {"error": error.error}

        query = urlencode(query_params)

        location = "%s?%s" % (self.client.redirect_uri, query)

        response.status_code = 302
        response.body = ""
        response.add_header("Location", location)

        return response

    def _generate_location(self, code):
        query = "code=" + code

        if self.state is not None:
            query += "&state=" + quote(self.state)

        return "%s?%s" % (self.client.redirect_uri, query)


class AuthorizationCodeTokenHandler(AccessTokenMixin, GrantHandler):
    """
    Implementation of the second step of the Authorization Code Grant
    (three-legged).
    """

    def __init__(self, auth_token_store, client_authenticator, **kwargs):
        self.client = None
        self.code = None
        self.data = {}
        self.redirect_uri = None
        self.scopes = []
        self.user_id = None

        self.auth_code_store = auth_token_store
        self.client_authenticator = client_authenticator

        super(AuthorizationCodeTokenHandler, self).__init__(**kwargs)

    def read_validate_params(self, request):
        """
        Reads and validates the data from the incoming request.

        A valid request is issued via POST consists of the following form-encoded body:

        client_id - Identifier of the requesting client (required)
        client_secret - Secret phrase generated by the auth system (required)
        code - Authorization code acquired in the Authorization Request (required)
        redirect_uri - URI that the OAuth2 server should redirect to (optional)
        """
        self._read_params(request)

        self._validate_code()

        return True

    def process(self, request, response, environ):
        """
        Generates a new access token and returns it.

        Returns the access token and the type of the token as JSON.

        Calls `oauth2.store.AccessTokenStore` to persist the token.
        """
        token_data = self.create_token(
            client_id=self.client.identifier,
            data=self.data,
            grant_type=AuthorizationCodeGrant.grant_type,
            scopes=self.scopes,
            user_id=self.user_id)

        self.auth_code_store.delete_code(self.code)

        if self.scopes:
            token_data["scope"] = encode_scopes(self.scopes)

        json_success_response(data=token_data, response=response)

        return response

    def handle_error(self, error, response):
        return json_error_response(error, response)

    def _read_params(self, request):
        self.client = self.client_authenticator.by_identifier_secret(request)
        self.code = request.post_param("code")
        self.redirect_uri = request.post_param("redirect_uri")

        if self.code is None or self.redirect_uri is None:
            raise OAuthInvalidError(
                error="invalid_request",
                explanation="Missing required parameter in request")

        try:
            self.client.redirect_uri = self.redirect_uri
        except RedirectUriUnknown:
            raise OAuthInvalidError(
                error="invalid_request",
                explanation="Invalid redirect_uri parameter")

    def _validate_code(self):
        try:
            stored_code = self.auth_code_store.fetch_by_code(self.code)
        except AuthCodeNotFound:
            raise OAuthInvalidError(
                error="invalid_request",
                explanation="Invalid authorization code parameter")

        if stored_code.code != self.code:
            raise OAuthInvalidError(
                error="invalid_grant",
                explanation="Invalid code parameter in request")

        if stored_code.redirect_uri != self.redirect_uri:
            raise OAuthInvalidError(
                error="invalid_request",
                explanation="Invalid redirect_uri parameter")

        if stored_code.is_expired():
            raise OAuthInvalidError(
                error="invalid_grant",
                explanation="Authorization code has expired")

        self.data = stored_code.data
        self.scopes = stored_code.scopes
        self.user_id = stored_code.user_id


class AuthorizationCodeGrant(GrantHandlerFactory, ScopeGrant,
                             SiteAdapterMixin):
    """
    Implementation of the Authorization Code Grant auth flow.

    This is a three-legged OAuth process.

    Register an instance of this class with
    :class:`oauth2.AuthorizationController` like this::

        auth_controller = AuthorizationController()

        auth_controller.add_grant_type(AuthorizationCodeGrant())

    .. versionchanged:: 1.0.0
       Require parameter ``site_adapter``.
    """

    grant_type = "authorization_code"

    site_adapter_class = AuthorizationCodeGrantSiteAdapter

    def __init__(self, unique_token=False, expires_in=0, **kwargs):
        self.unique_token = unique_token
        self.expires_in = expires_in

        super(AuthorizationCodeGrant, self).__init__(**kwargs)

    def __call__(self, request, server):
        """
        :param request: Incoming request
        :type request: oauth2.web.Request

        :param server: The OAuth2 provider instance


        .. versionchanged:: 1.0.0
           Check the HTTP method of a request
        """
        if (request.method == "POST"
            and request.post_param("grant_type") == "authorization_code"
            and request.path == server.token_path):
            return AuthorizationCodeTokenHandler(
                access_token_store=server.access_token_store,
                auth_token_store=server.auth_code_store,
                client_authenticator=server.client_authenticator,
                token_generator=server.token_generator,
                unique_token=self.unique_token)

        if (request.method == "GET"
            and request.get_param("response_type") == "code"
            and request.path == server.authorize_path):
            scope_handler = self._create_scope_handler()

            return AuthorizationCodeAuthHandler(
                auth_token_store=server.auth_code_store,
                client_authenticator=server.client_authenticator,
                scope_handler=scope_handler,
                site_adapter=self.site_adapter,
                token_generator=server.token_generator)

        return None


class ImplicitGrant(GrantHandlerFactory, ScopeGrant, SiteAdapterMixin):
    """
    Implementation of the Implicit Grant auth flow.

    This is a three-legged OAuth process.

    Register an instance of this class with
    :class:`oauth2.AuthorizationController` like this::

        auth_controller = AuthorizationController()

        auth_controller.add_grant_type(ImplicitGrant())

    .. versionchanged:: 1.0.0
       Require parameter ``site_adapter``.
    """

    grant_type = "implicit"

    site_adapter_class = ImplicitGrantSiteAdapter

    def __call__(self, request, server):
        response_type = request.get_param("response_type")

        if (response_type == "token"
            and request.path == server.authorize_path):
            return ImplicitGrantHandler(
                access_token_store=server.access_token_store,
                client_authenticator=server.client_authenticator,
                scope_handler=self._create_scope_handler(),
                site_adapter=self.site_adapter,
                token_generator=server.token_generator)
        return None


class ImplicitGrantHandler(AuthorizeMixin, AuthRequestMixin, GrantHandler):
    def __init__(self, access_token_store, **kwargs):
        self.access_token_store = access_token_store

        super(ImplicitGrantHandler, self).__init__(**kwargs)

    def process(self, request, response, environ):
        data = self.authorize(request, response, environ,
                              self.scope_handler.scopes)

        if isinstance(data, Response):
            return data

        token = self.token_generator.generate()

        access_token = AccessToken(client_id=self.client.identifier,
                                   grant_type=ImplicitGrant.grant_type,
                                   token=token, data=data[0],
                                   scopes=self.scope_handler.scopes)

        self.access_token_store.save_token(access_token)

        return self._redirect_access_token(response, token)

    def handle_error(self, error, response):
        redirect_location = "%s#error=%s" % (self.client.redirect_uri,
                                             error.error)

        response.add_header("Location", redirect_location)
        response.body = ""
        response.status_code = 302

        return response

    def _redirect_access_token(self, response, token):
        uri_with_fragment = "{0}#access_token={1}&token_type=bearer". \
            format(self.client.redirect_uri, token)

        if self.state is not None:
            uri_with_fragment += "&state=" + quote(self.state)

        if self.scope_handler.send_back is True:
            scopes_as_string = encode_scopes(self.scope_handler.scopes,
                                             use_quote=True)
            uri_with_fragment += "&scope=" + scopes_as_string

        response.status_code = 302
        response.add_header("Location", uri_with_fragment)
        response.content = ""

        return response


class ResourceOwnerGrant(GrantHandlerFactory, ScopeGrant, SiteAdapterMixin):
    """
    Implementation of the Resource Owner Password Credentials Grant auth flow.

    In this Grant a user provides a user name and a password.
    An access token is issued if the auth server was able to verify the user
    by her credentials.

    Register an instance of this class with
    :class:`oauth2.AuthorizationController` like this::

        auth_controller = AuthorizationController()

        auth_controller.add_grant_type(ResourceOwnerGrant())

    .. versionchanged:: 1.0.0
       Require parameter ``site_adapter``.
    """

    grant_type = "password"

    site_adapter_class = ResourceOwnerGrantSiteAdapter

    def __init__(self, unique_token=False, expires_in=0, **kwargs):
        self.unique_token = unique_token
        self.expires_in = expires_in

        super(ResourceOwnerGrant, self).__init__(**kwargs)

    def __call__(self, request, server):
        """
        Checks if the incoming request can be handled by the
        ResourceOwnerGrantHandler and returns an instance of it.
        """
        if request.post_param("grant_type") != self.grant_type:
            return None

        return ResourceOwnerGrantHandler(
            access_token_store=server.access_token_store,
            client_authenticator=server.client_authenticator,
            scope_handler=self._create_scope_handler(),
            site_adapter=self.site_adapter,
            token_generator=server.token_generator,
            unique_token=self.unique_token)


class ResourceOwnerGrantHandler(GrantHandler, AccessTokenMixin):
    """
    Class for handling Resource Owner authorization requests.

    See http://tools.ietf.org/html/rfc6749#section-4.3
    """

    OWNER_NOT_AUTHENTICATED = "Unable to authenticate resource owner"

    def __init__(self, client_authenticator, scope_handler, site_adapter,
                 **kwargs):
        """
        :param client_authenticator: Client authenticator
        :type client_authenticator: oauth2.client_authenticator.ClientAuthenticator

        :param scope_handler: Scope handler
        :type scope_handler: oauth2.grant.Scope

        :param site_adapter: Site adapter
        :type site_adapter: oauth2.web.SiteAdapter
        """
        self.client_authenticator = client_authenticator
        self.scope_handler = scope_handler
        self.site_adapter = site_adapter

        self.client = None
        self.password = None
        self.username = None

        super(ResourceOwnerGrantHandler, self).__init__(**kwargs)

    def process(self, request, response, environ):
        """
        Takes the incoming request, asks the concrete SiteAdapter to validate
        it and issues a new access token that is returned to the client on
        successful validation.
        """
        try:
            data = self.site_adapter.authenticate(request, environ,
                                                  self.scope_handler.scopes,
                                                  self.client)
            data = AuthorizeMixin.sanitize_return_value(data)
        except UserNotAuthenticated:
            raise OAuthInvalidError(error="invalid_client",
                                    explanation=self.OWNER_NOT_AUTHENTICATED)

        if isinstance(data, Response):
            return data

        token_data = self.create_token(
            client_id=self.client.identifier,
            data=data[0],
            grant_type=ResourceOwnerGrant.grant_type,
            scopes=self.scope_handler.scopes,
            user_id=data[1])

        if self.scope_handler.send_back:
            token_data["scope"] = encode_scopes(self.scope_handler.scopes)

        json_success_response(data=token_data, response=response)

        return response

    def read_validate_params(self, request):
        """
        Checks if all incoming parameters meet the expected values.
        """
        self.client = self.client_authenticator.by_identifier_secret(request)

        self.password = request.post_param("password")
        self.username = request.post_param("username")

        self.scope_handler.parse(request=request, source="body")

        return True

    def handle_error(self, error, response):
        status_code = 400
        if error.explanation == self.OWNER_NOT_AUTHENTICATED:
            status_code = 401

        return json_error_response(error, response, status_code=status_code)


class RefreshToken(GrantHandlerFactory, ScopeGrant):
    """
    Handles requests for refresh tokens as defined in
    http://tools.ietf.org/html/rfc6749#section-6.

    Adding a Refresh Token to the :class:`oauth2.AuthorizationController` like
    this::

        auth_controller = AuthorizationController()

        auth_controller.add_grant_type(ResourceOwnerGrant(tokens_expire=600))
        auth_controller.add_grant_type(RefreshToken(tokens_expire=1200))

    will cause :class:`oauth2.grant.AuthorizationCodeGrant` and
    :class:`oauth2.grant.ResourceOwnerGrant` to include a refresh token and
    expiration in the response.
    If tokens_expire == 0, the tokens will never expire.
    """

    grant_type = "refresh_token"

    def __init__(self, expires_in, reissue_refresh_tokens=False, **kwargs):

        self.refresh_expires_in = expires_in
        self.reissue_refresh_tokens = reissue_refresh_tokens

        super(RefreshToken, self).__init__(**kwargs)

    def __call__(self, request, server):
        """
        Determines if the current request requests a refresh token.

        :return: An instance of :class:`RefreshTokenHandler`.
        """
        if request.path != server.token_path:
            return None

        if request.post_param("grant_type") != "refresh_token":
            return None

        return RefreshTokenHandler(
            access_token_store=server.access_token_store,
            client_authenticator=server.client_authenticator,
            scope_handler=self._create_scope_handler(),
            token_generator=server.token_generator,
            reissue_refresh_tokens=self.reissue_refresh_tokens
        )


class RefreshTokenHandler(GrantHandler):
    """
    Validates an incoming request and issues a new access token.
    """

    def __init__(self, access_token_store, client_authenticator,
                 scope_handler, token_generator,
                 reissue_refresh_tokens=False):
        self.access_token_store = access_token_store
        self.client_authenticator = client_authenticator
        self.scope_handler = scope_handler
        self.token_generator = token_generator

        self.client = None
        self.data = {}
        self.refresh_grant_type = None
        self.refresh_token = None
        self.user_id = None

        self.reissue_refresh_tokens = reissue_refresh_tokens

    def process(self, request, response, environ):
        """
        Create a new access token.

        :param request: The incoming :class:`oauth2.web.Request`.
        :param response: The :class:`oauth2.web.Response` that will be returned
                         to the client.
        :param environ: A ``dict`` containing data of the environment.

        :return: :class:`oauth2.web.Response`

        """
        token_data = self.token_generator.create_access_token_data(self.refresh_grant_type)
        expires_at = int(time.time()) + token_data["expires_in"]

        access_token = AccessToken(client_id=self.client.identifier,
                                   token=token_data["access_token"],
                                   grant_type=self.refresh_grant_type,
                                   data=self.data, expires_at=expires_at,
                                   scopes=self.scope_handler.scopes,
                                   user_id=self.user_id)

        if self.reissue_refresh_tokens:
            self.access_token_store.delete_refresh_token(self.refresh_token)
            access_token.refresh_token = token_data["refresh_token"]
            refresh_expires_in = self.token_generator.refresh_expires_in
            refresh_expires_at = int(time.time()) + refresh_expires_in
            access_token.refresh_expires_at = refresh_expires_at
        else:
            del token_data["refresh_token"]

        self.access_token_store.save_token(access_token)

        json_success_response(data=token_data, response=response)

        return response

    def read_validate_params(self, request):
        """
        Validate the incoming request.

        :param request: The incoming :class:`oauth2.web.Request`.

        :return: Returns ``True`` if data is valid.

        :raises: :class:`oauth2.error.OAuthInvalidError`

        """
        self.refresh_token = request.post_param("refresh_token")

        if self.refresh_token is None:
            raise OAuthInvalidError(
                error="invalid_request",
                explanation="Missing refresh_token in request body")

        self.client = self.client_authenticator.by_identifier_secret(request)

        try:
            access_token = self.access_token_store.fetch_by_refresh_token(
                self.refresh_token
            )
        except AccessTokenNotFound:
            raise OAuthInvalidError(error="invalid_request",
                                    explanation="Invalid refresh token")

        refresh_token_expires_at = access_token.refresh_expires_at
        self.refresh_grant_type = access_token.grant_type

        if refresh_token_expires_at != 0 and \
                        refresh_token_expires_at < int(time.time()):
            raise OAuthInvalidError(error="invalid_request",
                                    explanation="Invalid refresh token")

        self.data = access_token.data
        self.user_id = access_token.user_id

        self.scope_handler.parse(request, "body")
        self.scope_handler.compare(access_token.scopes)

        return True

    def handle_error(self, error, response):
        return json_error_response(error, response)


class ClientCredentialsGrant(GrantHandlerFactory, ScopeGrant):
    grant_type = "client_credentials"

    def __call__(self, request, server):
        if request.path != server.token_path:
            return None

        if request.post_param("grant_type") == self.grant_type:
            return ClientCredentialsHandler(
                access_token_store=server.access_token_store,
                client_authenticator=server.client_authenticator,
                scope_handler=self._create_scope_handler(),
                token_generator=server.token_generator)
        return None


class ClientCredentialsHandler(GrantHandler):
    def __init__(self, access_token_store, client_authenticator,
                 scope_handler, token_generator):
        self.access_token_store = access_token_store
        self.client_authenticator = client_authenticator
        self.scope_handler = scope_handler
        self.token_generator = token_generator

        self.client = None

    def process(self, request, response, environ):
        body = {"token_type": "Bearer"}

        token = self.token_generator.generate()
        expires_in = self.token_generator.expires_in.get(ClientCredentialsGrant.grant_type, None)
        if expires_in is None:
            expires_at = None
        else:
            expires_at = int(time.time()) + expires_in

        access_token = AccessToken(
            client_id=self.client.identifier,
            grant_type=ClientCredentialsGrant.grant_type,
            token=token,
            expires_at=expires_at,
            scopes=self.scope_handler.scopes)
        self.access_token_store.save_token(access_token)

        body["access_token"] = token

        if expires_in is not None:
            body["expires_in"] = expires_in

        if self.scope_handler.send_back:
            body["scope"] = encode_scopes(self.scope_handler.scopes)

        json_success_response(data=body, response=response)

        return response

    def read_validate_params(self, request):
        self.client = self.client_authenticator.by_identifier_secret(request)

        self.scope_handler.parse(request=request, source="body")

    def handle_error(self, error, response):
        return json_error_response(error, response)
